"""
.. module:: radical.repex.md_kernels.amber_kernels_salt.amber_matrix_calculator_pattern_b
.. moduleauthor::  <haoyuan.chen@rutgers.edu>
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import json
import os,sys,socket,time

#-------------------------------------------------------------------------------
#
def reduced_energy(temperature, potential):
    """Calculates reduced energy.

    Arguments:
    temperature - replica temperature
    potential - replica potential energy

    Returns:
    reduced enery of replica
    """
    kb = 0.0019872041    #boltzmann const in kcal/mol
    if temperature != 0:
        beta = 1. / (kb*temperature)
    else:
        beta = 1. / kb     
    return float(beta * potential)

#-------------------------------------------------------------------------------
#
def get_historical_data(history_name, data_path=os.getcwd()):
    """Retrieves temperature and potential energy from simulation output file .history file.
    This file is generated after each simulation run. The function searches for directory 
    where .history file recides by checking all computeUnit directories on target resource.

    Arguments:
    history_name - name of .history file for a given replica. 

    Returns:
    data[0] - temperature obtained from .history file
    data[1] - potential energy obtained from .history file
    path_to_replica_folder - path to computeUnit directory on a target resource 
    where all
    input/output files for a given replica recide.
       Get temperature and potential energy from mdinfo file.

    ACTUALLY WE ONLY NEED THE POTENTIAL FROM HERE. TEMPERATURE GOTTA BE OBTAINED 
    FROM THE PROPERTY OF THE REPLICA OBJECT.
    """

    home_dir = os.getcwd()
    os.chdir(data_path)

    temp = 0.0    #temperature
    eptot = 0.0   #potential
    try:
        f = open(history_name)
        lines = f.readlines()
        f.close()
        path_to_replica_folder = os.getcwd()
        for i in range(len(lines)):
            #if "TEMP(K)" in lines[i]:
            #    temp = float(lines[i].split()[8])
            if "EPtot" in lines[i]:
                eptot = float(lines[i].split()[8])
    except:
        raise

    os.chdir(home_dir)
    return eptot, path_to_replica_folder

#-------------------------------------------------------------------------------
#
if __name__ == '__main__':
    """This module calculates one swap matrix column for replica and writes this column to 
    matrix_column_x_x.dat file. 
    """

    json_data = sys.argv[1]
    data=json.loads(json_data)

    replica_id = int(data["replica_id"])
    replica_cycle = int(data["replica_cycle"])
    replicas = int(data["replicas"])
    base_name = data["base_name"]

    prmtop_name = data["amber_parameters"]
    mdin_name = data["amber_input"]
    # INITIAL REPLICA TEMPERATURE:
    init_temp = float(data["init_temp"])
    init_salt = float(data["init_salt"])
    new_restraints = data["new_restraints"]

    # AMBER PATH ON THIS RESOURCE:
    amber_path = data["amber_path"]

    current_group_tsu = data["current_group_tsu"]

    r_old_path = data["r_old_path"]

    pwd = os.getcwd()

    #---------------------------------------------------------------------------

    temperatures = [0.0]*replicas  
    energies = [0.0]*replicas

    #for j in range(replicas):
    for j in current_group_tsu.keys():
        energy_history_name = base_name + "_" + str(j) + "_" + str(replica_cycle) + "_energy.mdinfo"
        try:
            rj_energy, path_to_replica_folder = get_historical_data( energy_history_name )
            temperatures[int(j)] = float(init_temp)
            energies[int(j)] = rj_energy
        except:
             raise

    # init swap column
    swap_column = [0.0]*replicas

    #for j in range(replicas):
    for j in current_group_tsu.keys():       
        swap_column[int(j)] = reduced_energy(temperatures[int(j)], energies[int(j)])

    #---------------------------------------------------------------------------
    # writing to file
    outfile = "matrix_column_{replica}_{cycle}.dat".format(cycle=replica_cycle, replica=replica_id )
    with open(outfile, 'w+') as f:
        row_str = ""
        for item in swap_column:        
            if len(row_str) != 0:
                row_str = row_str + " " + str(item)
            else:
                row_str = str(item)   
        row_str + " " + (str(path_to_replica_folder).rstrip())
        f.write(row_str)    
        f.write('\n')
        row_str = str(replica_id) + " " + str(replica_cycle) + " " + new_restraints + " " + str(init_temp) + " " + str(init_salt)
        f.write(row_str)
    f.close()
  
 


"""
.. module:: radical.repex.namd_kernels.namd_kernel_scheme_3
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
from os import path
from namd_kernel_tex import *

#-----------------------------------------------------------------------------------------------------------------------------------

class NamdKernelTexScheme3(NamdKernelTex):
    """This class is responsible for performing all operations related to NAMD for RE scheme 3.
    In this class is determined how replica input files are composed, how exchanges are performed, etc.

    RE scheme 3:
    - Asynchronous RE scheme: MD run on target resource is overlapped with local exchange step. Thus both MD run
    and exchange step are asynchronous.  
    - Number of replicas is greater than number of allocated resources.
    - Replica simulation cycle is defined by the fixed number of simulation time-steps each replica has to perform.
    - Exchange probabilities are determined using Gibbs sampling.
    - Exchange step is performed locally
    Overall algorithm is as follows:
        - First replicas in "waiting" state are submitted to pilot.
        - Then fixed time interval (cycle_time in input.json) must elapse before exchange step may take place.
        - After this fixed time interval elapsed, some replicas are still running on target resource.
        - In local exchange step are participating replicas which had finished MD run (state "finished") and
        replicas in "waiting" state.
        - After local exchanges step is performed replicas which participated in exchange are submitted to pilot
        to perform next simulation cycle
    """
    def __init__(self, inp_file,  work_dir_local):
        """Constructor.

        Arguments:
        inp_file - package input file with Pilot and NAMD related parameters as specified by user 
        work_dir_local - directory from which main simulation script was invoked
        """

        NamdKernelTex.__init__(self, inp_file, work_dir_local)

#-----------------------------------------------------------------------------------------------------------------------------------

    def check_replicas(self, replicas):
        """
        """
        finished_replicas = []
        files = os.listdir( self.work_dir_local )

        for r in replicas:
            history_name =  self.inp_basename[:-5] + "_%s_%s.history" % ( r.id, (r.cycle-1) )
            for item in files:
                if (item.startswith(history_name)):
                    if r not in finished_replicas:
                        finished_replicas.append( r )

        return finished_replicas





__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import math
import json
import time
import random

#-------------------------------------------------------------------------------
#
def weighted_choice_sub(weights):
    """Adopted from asyncre-bigjob [1]
    """

    rnd = random.random() * sum(weights)
    for i, w in enumerate(weights):
        rnd -= w
        if rnd < 0:
            return i

#-------------------------------------------------------------------------------
#
def gibbs_exchange(r_i, replicas, swap_matrix):
    """Adopted from asyncre-bigjob [1]
    Produces a replica "j" to exchange with the given replica "i"
    based off independence sampling of the discrete distribution

    Arguments:
    r_i - given replica for which is found partner replica
    replicas - list of Replica objects
    swap_matrix - matrix of dimension-less energies, where each column is a replica 
    and each row is a state

    Returns:
    r_j - replica to exchnage parameters with
    """

    #evaluate all i-j swap probabilities
    ps = [0.0]*(len(replicas))

    j = 0
    for r_j in replicas:
        ps[j] = -(swap_matrix[r_i.sid][r_j.id] + swap_matrix[r_j.sid][r_i.id] - 
                  swap_matrix[r_i.sid][r_i.id] - swap_matrix[r_j.sid][r_j.id]) 
        j += 1
        
    ######################################
    new_ps = []
    for item in ps:
        if item > math.log(sys.float_info.max): new_item=sys.float_info.max
        elif item < math.log(sys.float_info.min) : new_item=0.0
        else :
            new_item = math.exp(item)
        new_ps.append(new_item)
    ps = new_ps
    # index of swap replica within replicas_waiting list
    j = len(replicas)
    while j > (len(replicas) - 1):
        j = weighted_choice_sub(ps)
        
    # guard for errors
    if j is None:
        j = random.randint(0,(len(replicas)-1))
        print "...gibbs exchnage warning: j was None..."
    # actual replica
    r_j = replicas[j]
    ######################################

    return r_j

#-------------------------------------------------------------------------------
#
def do_exchange(dimension, replicas, swap_matrix):
    """
    """

    exchanged = []
    for r_i in replicas:
        r_j = gibbs_exchange(r_i, replicas, swap_matrix)
       
        if (r_j.id != r_i.id) and (r_j.id not in exchanged) and (r_i.id not in exchanged):
            exchanged.append(r_j.id)
            exchanged.append(r_i.id)
            
    return  exchanged

#-------------------------------------------------------------------------------
#
class Replica(object):
    """Class representing replica and it's associated data.
    """
    def __init__(self, my_id, new_temperature=None):
       
        self.id = int(my_id)
        self.sid = int(my_id)
        self.new_temperature = new_temperature

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    """
    """

    #argument_list = str(sys.argv)
    #replicas = int(sys.argv[1])
    #current_cycle = int(sys.argv[2])
    #dimension = int(sys.argv[3])

    argument_list = str(sys.argv)
    current_cycle = int(sys.argv[1])
    replicas = int(sys.argv[2])
    base_name = str(sys.argv[3])

    replica_dict = {}
    replicas_obj = []

    base_name = "matrix_column"

    # init matrix
    swap_matrix = [[ 0. for j in range(replicas)] for i in range(replicas)]

    for rid in range(replicas):
        success = 0
        column_file = base_name + "_" + str(rid) + "_" + str(current_cycle) + ".dat" 
        path = "../staging_area/" + column_file     
        while (success == 0):
            try:
                f = open(path)
                lines = f.readlines()
                f.close()
                #---------------------------------------------------------------
                # populating matrix column
                data = lines[0].split()
                for i in range(replicas):
                    swap_matrix[i][int(rid)] = float(data[i])
                #---------------------------------------------------------------
                # populating replica dict
                data = lines[1].split()
                # data[0] = rid; data[1] = cycle; data[2] = init_temp
                replica_dict[data[0]] = [data[1], data[2]]
                #---------------------------------------------------------------
                # creating replica
                r = Replica(rid, replica_dict[str(rid)][1])
                replicas_obj.append(r)
                success = 1
                print "Success processing replica: %s" % rid
            except:
                print "Waiting for replica: %s" % rid
                time.sleep(1)
                pass
    #---------------------------------------------------------------------------
    exchange_list = []
    for r_i in replicas_obj:
        r_j = gibbs_exchange(r_i, replicas_obj, swap_matrix)
        if (r_j != r_i):
            exchange_pair = []
            exchange_pair.append(r_i.id)
            exchange_pair.append(r_j.id)
            exchange_list.append(exchange_pair)
        
    #---------------------------------------------------------------------------
    # writing to file
    try:
        outfile = "pairs_for_exchange_{cycle}.dat".format(cycle=current_cycle)
        with open(outfile, 'w+') as f:
            for pair in exchange_list:
                if pair:
                    row_str = str(pair[0]) + " " + str(pair[1]) 
                    f.write(row_str)
                    f.write('\n')
        f.close()

    except IOError:
        print 'Error: unable to create column file %s for replica %s' % \
        (outfile, replica_id)




"""
.. module:: radical.repex.namd_kernels.namd_kernel_scheme_2
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
from os import path
import radical.pilot
from kernels.kernels import KERNELS
from namd_kernel_tex import *
import namd_kernels_tex.namd_matrix_calculator_scheme_2
import radical.utils.logger as rul

#-----------------------------------------------------------------------------------------------------------------------------------

class NamdKernelTexScheme2(NamdKernelTex):
    """This class is responsible for performing all operations related to NAMD for RE scheme 2.
    In this class is determined how replica input files are composed, how exchanges are performed, etc.

    RE scheme 2:
    - Synchronous RE scheme: none of the replicas can start exchange before all replicas has finished MD run.
    Conversely, none of the replicas can start MD run before all replicas has finished exchange step. 
    In other words global barrier is present.   
    - Number of replicas is greater than number of allocated resources for both MD and exchange step.
    - Simulation cycle is defined by the fixed number of simulation time-steps for each replica.
    - Exchange probabilities are determined using Gibbs sampling.
    - Exchange step is performed in decentralized fashion on target resource.

    """
    def __init__(self, inp_file, work_dir_local):
        """Constructor.

        Arguments:
        inp_file - package input file with Pilot and NAMD related parameters as specified by user 
        work_dir_local - directory from which main simulation script was invoked
        """

        self.shared_urls = []
        self.shared_files = []

        self.name = 'nk-patternB-tex'
        self.logger  = rul.getLogger ('radical.repex', self.name)

        NamdKernelTex.__init__(self, inp_file, work_dir_local)


#----------------------------------------------------------------------------------------------------------------------------------

    # ------------------------------------------------------------------------------
    #
    def prepare_shared_data(self):
 
        structure_path = self.work_dir_local + "/" + self.inp_folder + "/" + self.namd_structure
        coords_path = self.work_dir_local + "/" + self.inp_folder + "/" + self.namd_coordinates
        params_path = self.work_dir_local + "/" + self.inp_folder + "/" + self.namd_parameters

        self.shared_files.append(self.namd_structure)
        self.shared_files.append(self.namd_coordinates)
        self.shared_files.append(self.namd_parameters)

        struct_url = 'file://%s' % (structure_path)
        self.shared_urls.append(struct_url)
 
        coords_url = 'file://%s' % (coords_path)
        self.shared_urls.append(coords_url)     

        params_url = 'file://%s' % (params_path)
        self.shared_urls.append(params_url)
    
    # ------------------------------------------------------------------------------
    #
    def build_input_file(self, replica):
        """Generates input file for individual replica, based on template input file. Tokens @xxx@ are
        substituted with corresponding parameters. 

        old_output_root @oldname@ determines which .coor .vel and .xsc files are used for next cycle

        Arguments:
        replica - a single Replica object
        """

        basename = self.inp_basename[:-5]
        template = self.inp_basename
            
        new_input_file = "%s_%d_%d.namd" % (basename, replica.id, replica.cycle)
        outputname = "%s_%d_%d" % (basename, replica.id, replica.cycle)
        old_name = "%s_%d_%d" % (basename, replica.id, (replica.cycle-1))
        replica.new_coor = outputname + ".coor"
        replica.new_vel = outputname + ".vel"
        replica.new_history = outputname + ".history"
        replica.new_ext_system = outputname + ".xsc" 
        historyname = replica.new_history

        if (replica.cycle == 0):
            first_step = 0
        elif (replica.cycle == 1):
            first_step = int(self.cycle_steps)
        else:
            first_step = (replica.cycle - 1) * int(self.cycle_steps)

        if (replica.cycle == 0):
            old_name = "%s_%d_%d" % (basename, replica.id, (replica.cycle-1)) 
            structure = self.namd_structure
            coordinates = self.namd_coordinates
            parameters = self.namd_parameters
        else:
            old_name = replica.old_path + "/%s_%d_%d" % (basename, replica.id, (replica.cycle-1))
            structure = replica.first_path + "/" + self.namd_structure
            coordinates = replica.first_path + "/" + self.namd_coordinates
            parameters = replica.first_path + "/" + self.namd_parameters

        # substituting tokens in main replica input file 
        try:
            r_file = open( (os.path.join((self.work_dir_local + "/namd_inp/"), template)), "r")
        except IOError:
            print 'Warning: unable to access template file %s' % template

        tbuffer = r_file.read()
        r_file.close()

        tbuffer = tbuffer.replace("@swap@",str(replica.swap))
        tbuffer = tbuffer.replace("@ot@",str(replica.old_temperature))
        tbuffer = tbuffer.replace("@nt@",str(replica.new_temperature))
        tbuffer = tbuffer.replace("@steps@",str(self.cycle_steps))
        tbuffer = tbuffer.replace("@rid@",str(replica.id))
        tbuffer = tbuffer.replace("@somename@",str(outputname))
        tbuffer = tbuffer.replace("@oldname@",str(old_name))
        tbuffer = tbuffer.replace("@cycle@",str(replica.cycle))
        tbuffer = tbuffer.replace("@firststep@",str(first_step))
        tbuffer = tbuffer.replace("@history@",str(historyname))
        tbuffer = tbuffer.replace("@structure@", str(structure))
        tbuffer = tbuffer.replace("@coordinates@", str(coordinates))
        tbuffer = tbuffer.replace("@parameters@", str(parameters))
        
        replica.cycle += 1
        # write out
        try:
            w_file = open( new_input_file, "w")
            w_file.write(tbuffer)
            w_file.close()
        except IOError:
            print 'Warning: unable to access file %s' % new_input_file

#-----------------------------------------------------------------------------------------------------------------------------------

    def prepare_replica_for_md(self, replica, sd_shared_list):
        """Creates a list of ComputeUnitDescription objects for MD simulation step. Here are
        specified input/output files to be transferred to/from target resource. Note: input 
        files for first and subsequent simulaition cycles are different.  

        Arguments:
        replicas - list of Replica objects

        Returns:
        compute_replicas - list of radical.pilot.ComputeUnitDescription objects
        """
        
        self.build_input_file(replica)
        input_file = "%s_%d_%d.namd" % (self.inp_basename[:-5], replica.id, (replica.cycle-1))

        new_coor = replica.new_coor
        new_vel = replica.new_vel
        new_history = replica.new_history
        new_ext_system = replica.new_ext_system

        old_coor = replica.old_coor
        old_vel = replica.old_vel
        old_ext_system = replica.old_ext_system 

        st_out = []
 
        history_out = {
            'source': new_history,
            'target': 'staging:///%s' % new_history,
            'action': radical.pilot.COPY
        }
        st_out.append(history_out)
        
        coor_out = {
            'source': new_coor,
            'target': 'staging:///%s' % new_coor,
            'action': radical.pilot.COPY
        }                   
        st_out.append(coor_out)        

        vel_out = {
            'source': new_vel,
            'target': 'staging:///%s' % new_vel,
            'action': radical.pilot.COPY
        }
        st_out.append(vel_out)
        
        ext_out = {
            'source': new_ext_system,
            'target': 'staging:///%s' % new_ext_system,
            'action': radical.pilot.COPY
        }
        st_out.append(ext_out)

        # only for first cycle we transfer structure, coordinates and parameters files
        if replica.cycle == 1:
            cu = radical.pilot.ComputeUnitDescription()
            cu.pre_exec    = KERNELS[self.resource]["kernels"]["namd"]["pre_execution"]
            cu.executable = self.namd_path
            cu.arguments = [input_file]
            cu.cores = replica.cores
            cu.mpi = False
            cu.input_staging = [str(input_file)] + sd_shared_list
            cu.output_staging = st_out
            
        else:
            cu = radical.pilot.ComputeUnitDescription()
            cu.pre_exec    = KERNELS[self.resource]["kernels"]["namd"]["pre_execution"]
            cu.executable = self.namd_path
            cu.arguments = [input_file]
            cu.cores = replica.cores
            cu.mpi = False
            cu.input_staging = [str(input_file)] + sd_shared_list
            cu.output_staging = st_out

        return cu

#-----------------------------------------------------------------------------------------------------------------------------------

    def prepare_replica_for_exchange(self, replica, sd_shared_list):
        """Creates a list of ComputeUnitDescription objects for exchange step on resource.
        Number of matrix_calculator_scheme_2.py instances invoked on resource is equal to the number 
        of replicas. 

        Arguments:
        replicas - list of Replica objects

        Returns:
        exchange_replicas - list of radical.pilot.ComputeUnitDescription objects
        """
           
        # name of the file which contains swap matrix column data for each replica
        matrix_col = "matrix_column_%s_%s.dat" % (replica.id, (replica.cycle-1))
        basename = self.inp_basename[:-5]
        cu = radical.pilot.ComputeUnitDescription()
        cu.executable = "python"
        # each scheme has it's own calculator!
        calculator_path = os.path.dirname(namd_kernels_tex.namd_matrix_calculator_scheme_2.__file__)
        calculator = calculator_path + "/namd_matrix_calculator_scheme_2.py"
        cu.input_staging = [calculator] + sd_shared_list
        cu.arguments = ["namd_matrix_calculator_scheme_2.py", replica.id, (replica.cycle-1), self.replicas, basename]
        cu.cores = 1            
            
        return cu
            




__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import json
import math
import time
import shutil

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    """
    """

    json_data = sys.argv[1]
    data=json.loads(json_data)

    cycle_steps     = data["cycle_steps"]
    new_restraints  = data["new_restraints"]
    new_temperature = data["new_temperature"]
    amber_input     = data["amber_input"]
    new_input_file  = data["new_input_file"]
    us_template     = data["us_template"]
    replica_cycle   = int(data["cycle"])
    rstr_val_1      = float(data["rstr_val_1"])
    rstr_val_2      = float(data["rstr_val_2"])

    #---------------------------------------------------------------------------
    # this is for every cycle
    try:
        r_file = open(amber_input, "r")
    except IOError:
        print "Warning: unable to access template file: {0}".format(amber_input) 

    tbuffer = r_file.read()
    r_file.close()

    tbuffer = tbuffer.replace("@nstlim@",cycle_steps)
    tbuffer = tbuffer.replace("@disang@",new_restraints)
    tbuffer = tbuffer.replace("@temp@",new_temperature)

    if replica_cycle == 1:
        tbuffer = tbuffer.replace("@irest@","0")
        tbuffer = tbuffer.replace("@ntx@","1")
    else:
        tbuffer = tbuffer.replace("@irest@","1")
        tbuffer = tbuffer.replace("@ntx@","5")

    try:
        w_file = open(new_input_file, "w")
        w_file.write(tbuffer)
        w_file.close()
    except IOError:
        print "Warning: unable to access file: {0}".format(new_input_file)

    #---------------------------------------------------------------------------
    # this is for first cycle only
    if replica_cycle == 1:
        print "first cycle"
        try:
            r_file = open(us_template, "r")
            tbuffer = r_file.read()
            r_file.close()
        except IOError:
            print "Warning: unable to access file: {0}".format(us_template) 

        try:
            w_file = open(new_restraints, "w")
            tbuffer = tbuffer.replace("@val1@", str(rstr_val_1))
            tbuffer = tbuffer.replace("@val1l@", str(rstr_val_1-90))
            tbuffer = tbuffer.replace("@val1h@", str(rstr_val_1+90))
            tbuffer = tbuffer.replace("@val2@", str(rstr_val_2))
            tbuffer = tbuffer.replace("@val2l@", str(rstr_val_2-90))
            tbuffer = tbuffer.replace("@val2h@", str(rstr_val_2+90))
            w_file.write(tbuffer)
            w_file.close()
        except IOError:
            print "Warning: unable to access file: {0}".format(new_restraints)
 
        #-----------------------------------------------------------------------
        # copy to staging area
        pwd = os.getcwd()
        src = pwd + "/" + new_restraints

        size = len(pwd)-1
        path = pwd
        for i in range(0,size):
            if pwd[size-i] != '/':
                path = path[:-1]
            else:
                break

        dst = path + "staging_area/" + new_restraints 
        shutil.copyfile(src, dst)




"""
.. module:: radical.repex.namd_kernels.matrix_calculator_scheme_2
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys

#-----------------------------------------------------------------------------------------------------------------------------------

def reduced_energy(temperature, potential):
    """Calculates reduced energy.

    Arguments:
    temperature - replica temperature
    potential - replica potential energy

    Returns:
    reduced enery of replica
    """
    kb = 0.0019872041
    # check for division by zero
    if temperature != 0:
        beta = 1. / (kb*temperature)
    else:
        beta = 1. / kb     
    return float(beta * potential)

#-----------------------------------------------------------------------------------------------------------------------------------

def get_historical_data(history_name):
    """Retrieves temperature and potential energy from simulation output file .history file.
    This file is generated after each simulation run. The function searches for directory 
    where .history file recides by checking all computeUnit directories on target resource.

    Arguments:
    history_name - name of .history file for a given replica. 

    Returns:
    data[0] - temperature obtained from .history file
    data[1] - potential energy obtained from .history file
    path_to_replica_folder - path to computeUnit directory on a target resource where all
    input/output files for a given replica recide.
    """
    home_dir = os.getcwd()
    os.chdir("../")
    
    os.chdir("staging_area")
    try:
        f = open(history_name)
        lines = f.readlines()
        f.close()
        path_to_replica_folder = os.getcwd()
        data = lines[0].split()
    except:
        pass 
    os.chdir("../")
 
    os.chdir(home_dir)
    return float(data[0]), float(data[1]), path_to_replica_folder


#-----------------------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':
    """This module calculates one swap matrix column for replica and writes this column to 
    matrix_column_x_x.dat file. 
    """

    argument_list = str(sys.argv)
    replica_id = str(sys.argv[1])
    replica_cycle = str(sys.argv[2])
    replicas = int(str(sys.argv[3]))
    base_name = str(sys.argv[4])

    pwd = os.getcwd()

    # getting history data for self
    history_name = base_name + "_" + replica_id + "_" + replica_cycle + ".history"
    replica_temp, replica_energy, path_to_replica_folder = get_historical_data( history_name )

    # getting history data for all replicas
    # we rely on the fact that last cycle for every replica is the same, e.g. == replica_cycle
    # but this is easily changeble for arbitrary cycle numbers
    temperatures = [0.0]*replicas
    energies = [0.0]*replicas
    for j in range(replicas):
        history_name = base_name + "_" + str(j) + "_" + replica_cycle + ".history" 
        try:
            rj_temp, rj_energy, temp = get_historical_data( history_name )
            temperatures[j] = rj_temp
            energies[j] = rj_energy
        except:
             pass 

    # init swap column
    swap_column = [0.0]*replicas

    for j in range(replicas):        
        swap_column[j] = reduced_energy(temperatures[j], replica_energy)


    for item in swap_column:
        print item,    
        
    # printing path
    print str(path_to_replica_folder).rstrip()

    print str(replica_id)
    



"""
.. module:: radical.repex.replicas.replica
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import math
import json

#-------------------------------------------------------------------------------

class Replica1d(object):
    """Class representing replica and it's associated data.
    """
    def __init__(self, my_id, new_temperature=None, new_salt_concentration=None, new_restraints=None, rstr_val_1=None,  coor=None, indx1=None, indx2=None, cores=1):
        """Constructor.

        Arguments:
        my_id - integer representing replica's id
        new_temperature - temperature at which replica is initialized (default: None)
        cores - number of cores to be used by replica's NAMD instance (default: 1)
        """
        self.id = int(my_id)
        self.sid = int(my_id)
        self.state = 'I'
        self.cycle = 0

        if rstr_val_1 is None:
            self.rstr_val_1 = 0
        else:
            self.rstr_val_1 = rstr_val_1

        if coor is not None:
            self.coor_file = coor

        if indx1 is not None:
            self.indx1 = indx1

        if indx2 is not None:
            self.indx2 = indx2

        #---------------------------------------------------------
        if new_temperature is None:
            self.new_temperature = 0
        else:
            self.new_temperature = new_temperature
        self.old_temperature = new_temperature
        
        #---------------------------------------------------------
        if new_salt_concentration is None:
            self.new_salt_concentration = 0
        else:
            self.new_salt_concentration = new_salt_concentration
        self.old_salt_concentration = new_salt_concentration
        #---------------------------------------------------------

        if new_restraints is None:
            self.new_restraints = ''
        else:
            self.new_restraints = new_restraints
        self.old_restraints = new_restraints
        #---------------------------------------------------------

        self.potential = 0

        # amber stuff
        self.new_traj = ""
        self.new_info = ""
        self.old_traj = ""
        self.old_info = ""
        ###################

        self.new_coor = ""
        self.new_vel = ""
        self.new_history = ""
        self.new_ext_system = "" 
        self.old_coor = ""
        self.old_vel = ""
        self.old_ext_system = "" 
        self.old_path = ""
        self.first_path = ""
        self.swap = 0
        self.cores = cores
        self.stopped_run = -1

#-------------------------------------------------------------------------------

class Replica2d(object):
    """Class representing replica and it's associated data.
    """
    def __init__(self, my_id, new_temperature=None, new_salt_concentration=None, cores=1):
        """Constructor.

        """
        self.id = int(my_id)
        self.sid = int(my_id)
        self.cycle = 0
        if new_salt_concentration is None:
            self.new_salt_concentration = 0
        else:
            self.new_salt_concentration = new_salt_concentration
        self.old_salt_concentration = new_salt_concentration

        if new_temperature is None:
            self.new_temperature = 0
        else:
            self.new_temperature = new_temperature
        self.old_temperature = new_temperature

        self.potential = 0

        # amber stuff
        self.new_traj = ""  # ok
        self.new_info = ""  # ok
        self.new_coor = ""  # ok both namd and amber


        self.old_traj = ""  # ok
        self.old_info = ""  # ok 
        self.old_coor = ""  # ok both namd and amber
        ###################

        
        self.new_vel = ""         # namd only
        self.new_history = ""     # namd only
        self.new_ext_system = ""  # namd only
        self.old_vel = ""         # namd only
        self.old_ext_system = ""  # namd only


        self.old_path = ""
        self.first_path = ""
        self.swap = 0
        self.cores = cores
        self.stopped_run = -1

#-------------------------------------------------------------------------------

class Replica3d(object):
    """Class representing replica and it's associated data.
       US = Umbrella Sampling
    """
    def __init__(self, my_id, new_temperature=None, new_salt=None, new_restraints=None, rstr_val_1=None, rstr_val_2=None, coor=None, indx1=None, indx2=None, cores=1):
        """Constructor.

        Arguments:
        my_id - integer representing replica's id
        new_temperature - temperature at which replica is initialized (default: None)
        cores - number of cores to be used by replica's NAMD instance (default: 1)
        """
        self.id = int(my_id)
        self.sid = int(my_id)
        self.state = 'I'
        self.cycle = 0

        if coor is not None:
            self.coor_file = coor

        if indx1 is not None:
            self.indx1 = indx1

        if indx2 is not None:
            self.indx2 = indx2

        if new_salt is None:
            self.new_salt_concentration = 0
        else:
            self.new_salt_concentration = new_salt
        self.old_salt_concentration = new_salt

        if new_restraints is None:
            self.new_restraints = ''
            self.old_restraints = ''
        else:
            self.new_restraints = new_restraints
            self.old_restraints = new_restraints
        self.potential_1 = 0

        if new_temperature is None:
            self.new_temperature = 0
        else:
            self.new_temperature = new_temperature
        self.old_temperature = new_temperature

        if rstr_val_1 is None:
            self.rstr_val_1 = 0
        else:
            self.rstr_val_1 = rstr_val_1

        if rstr_val_2 is None:
            self.rstr_val_2 = 0
        else:
            self.rstr_val_2 = rstr_val_2

        # amber stuff
        self.new_traj = ""  # ok
        self.new_info = ""  # ok
        self.new_coor = ""  # ok both namd and amber


        self.old_traj = ""  # ok
        self.old_info = ""  # ok 
        self.old_coor = ""  # ok both namd and amber
        ###################
        ###################

        self.old_path = ""
        self.first_path = ""
        self.swap = 0
        self.cores = cores
        self.stopped_run = -1


"""
.. module:: radical.repex.namd_kernels.namd_kernel_scheme_2a
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

from namd_kernel_tex import *

#-----------------------------------------------------------------------------------------------------------------------------------

class NamdKernelTexScheme2a(NamdKernelTex):
    """This class is responsible for performing all operations related to NAMD for RE scheme 2a.
    In this class is determined how replica input files are composed, how exchanges are performed, etc.

    RE scheme 2a:
    - Synchronous RE scheme: none of the replicas can start exchange before all replicas has finished MD run.
    Conversely, none of the replicas can start MD run before all replicas has finished exchange step. 
    In other words global barrier is present.   
    - Number of replicas is greater than number of allocated resources for both MD and exchange step.
    - Simulation cycle is defined by the fixed number of simulation time-steps for each replica.
    - Exchange probabilities are determined using Gibbs sampling.
    - Exchange step is performed in decentralized fashion on target resource.

    """
    def __init__(self, inp_file,  work_dir_local):
        """Constructor.

        Arguments:
        inp_file - package input file with Pilot and NAMD related parameters as specified by user 
        work_dir_local - directory from which main simulation script was invoked
        """

        NamdKernelTex.__init__(self, inp_file, work_dir_local)




"""
.. module:: radical.repex.md_kernles.amber_kernels_us.kernel_pattern_s_us
.. moduleauthor::  <antons.treikalis@rutgers.edu>
.. moduleauthor::  <haoyuan.chen@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import time
import math
import json
import random
import shutil
import datetime
from os import path
import radical.pilot
from os import listdir
from os.path import isfile, join
import radical.utils.logger as rul
from replicas.replica import Replica1d
from kernels.kernels import KERNELS
import amber_kernels_us.input_file_builder
import amber_kernels_us.global_ex_calculator
#import amber_kernels_us.global_ex_calculator_mpi
import amber_kernels_us.ind_ex_calculator

#-------------------------------------------------------------------------------

class KernelPatternSus(object):
    """

    """
    def __init__(self, inp_file, rconfig, work_dir_local):
        """Constructor.

        Arguments:
        inp_file - package input file with Pilot and NAMD related parameters as 
        specified by user 
        work_dir_local - directory from which main simulation script was invoked
        """

        self.resource = rconfig['target']['resource']
        self.inp_basename = inp_file['remd.input']['input_file_basename']
        self.input_folder = inp_file['remd.input']['input_folder']
        self.replicas = int(inp_file['remd.input']['number_of_replicas'])
        self.cores = int(rconfig['target']['cores'])
        self.us_template = inp_file['remd.input']['us_template']                           
        self.init_temperature = float(inp_file['remd.input']['init_temperature'])
        self.cycle_steps = int(inp_file['remd.input']['steps_per_cycle'])
        self.work_dir_local = work_dir_local
        
        try:
            self.nr_cycles = int(inp_file['remd.input']['number_of_cycles'])
        except:
            self.nr_cycles = None

        if 'replica_mpi' in inp_file['remd.input']:
            mpi = inp_file['remd.input']['replica_mpi']
            if mpi == "True":
                self.replica_mpi = True
            else:
                self.replica_mpi = False
        else:
            self.replica_mpi = False

        if 'exchange_mpi' in inp_file['remd.input']:
            mpi = inp_file['remd.input']['exchange_mpi']
            if mpi == "True":
                self.exchange_mpi = True
            else:
                self.exchange_mpi = False
        else:
            self.exchange_mpi = False
            
        if 'download_mdinfo' in inp_file['remd.input']:
            if inp_file['remd.input']['download_mdinfo'] == 'True':
                self.down_mdinfo = True
            else:
                self.down_mdinfo = False
        else:
            self.down_mdinfo = True

        if 'download_mdout' in inp_file['remd.input']:
            if inp_file['remd.input']['download_mdout'] == 'True':
                self.down_mdout = True
            else:
                self.down_mdout = False
        else:
            self.down_mdout = True

        try:
            self.replica_cores = inp_file['remd.input']['replica_cores']
        except:
            self.replica_cores = 1

        self.restraints_files = []
        for k in range(self.replicas):
            self.restraints_files.append(self.us_template + "." + str(k) )

        self.pre_exec = KERNELS[self.resource]["kernels"]["amber"]["pre_execution"]
        try:
            self.amber_path = inp_file['remd.input']['amber_path']
        except:
            print "Using default Amber path for %s" % rconfig['target']['resource']
            try:
                self.amber_path = KERNELS[self.resource]["kernels"]["amber"]["executable"]
            except:
                print "Amber path for localhost is not defined..."

        self.amber_coordinates_path = inp_file['remd.input']['amber_coordinates_folder']
        if 'same_coordinates' in inp_file['remd.input']:
            coors = inp_file['remd.input']['same_coordinates']
            if coors == "True":
                self.same_coordinates = True
            else:
                self.same_coordinates = False
        else:
            self.same_coordinates = True

        self.amber_parameters = inp_file['remd.input']['amber_parameters']
        self.amber_input = inp_file['remd.input']['amber_input']
        self.input_folder = inp_file['remd.input']['input_folder']
        self.us_start_param = float(inp_file['remd.input']['us_start_param'])
        self.us_end_param = float(inp_file['remd.input']['us_end_param'])

        self.current_cycle = -1

        self.name = 'ak-us-patternB'
        self.ex_name = 'umbrella_sampling'
        self.logger  = rul.getLogger ('radical.repex', self.name)

        self.shared_urls = []
        self.shared_files = []

    #---------------------------------------------------------------------------
    #
    def initialize_replicas(self):
        """Initializes replicas and their attributes to default values

           Changed to use geometrical progression for temperature assignment.

           The user needs to provide the indexed restraint files now. A script 
           that generates those files will be provided later.
        """
        replicas = []

        #-----------------------------------------------------------------------
        # parse coor file
        coor_path  = self.work_dir_local + "/" + self.input_folder + "/" + self.amber_coordinates_path
        coor_list  = listdir(coor_path)

        base = coor_list[0]

        self.c_prefix = ''
        self.c_infix = ''
        self.c_postfix = ''

        under = 0
        dot = 0
        for ch in base:
            if ch == '_':
                under += 1
            if ch == '.':
                dot += 1

            if under == 0 and dot == 0:
                self.c_prefix += ch
            elif under == 1 and dot == 1 and ch != '.':
                self.c_infix += ch
            elif under == 2 and dot == 2 and ch != '.':
                self.c_postfix += ch

        #-----------------------------------------------------------------------
        # 
        for k in range(self.replicas):

            spacing = (self.us_end_param - self.us_start_param) / (float(self.replicas)-1)
            starting_value = self.us_start_param + k*spacing

            rstr_val_1 = str(starting_value)

            if self.same_coordinates == False:
                coor_file = self.c_prefix + "_0." + self.c_infix + "_" + str(k) + "." + self.c_postfix
                r = Replica1d(k, new_restraints=self.restraints_files[k], \
                             rstr_val_1=float(rstr_val_1), \
                             coor=coor_file, \
                             indx1=0, \
                             indx2=k)
            else:
                coor_file = self.c_prefix + "_0." + self.c_infix + "_0." + self.c_postfix
                r = Replica1d(k, new_restraints=self.restraints_files[k], \
                                 rstr_val_1=float(rstr_val_1), \
                                 coor=coor_file)

            replicas.append(r)
            
        return replicas

    #---------------------------------------------------------------------------
    # 
    def prepare_shared_data(self):

        parm_path = self.work_dir_local + "/" + self.input_folder + "/" + self.amber_parameters
        coor_path  = self.work_dir_local + "/" + self.input_folder + "/" + self.amber_coordinates_path
        inp_path  = self.work_dir_local + "/" + self.input_folder + "/" + self.amber_input

        build_inp = os.path.dirname(amber_kernels_us.input_file_builder.__file__)
        build_inp_path = build_inp + "/input_file_builder.py"

        global_calc = os.path.dirname(amber_kernels_us.global_ex_calculator.__file__)
        global_calc_path = global_calc + "/global_ex_calculator.py"

        rstr_template_path = self.work_dir_local + "/" + self.input_folder + "/" + self.us_template

        global_calc_s = os.path.dirname(amber_kernels_us.global_ex_calculator.__file__)
        global_calc_path_s = global_calc_s + "/global_ex_calculator.py"

        ind_calc = os.path.dirname(amber_kernels_us.ind_ex_calculator.__file__)
        ind_calc_path = ind_calc + "/ind_ex_calculator.py"
        #-----------------------------------------------------------------------

        self.shared_files.append(self.amber_parameters)
        self.shared_files.append(self.amber_input)
        self.shared_files.append("input_file_builder.py")
        self.shared_files.append("global_ex_calculator.py")
        self.shared_files.append(self.us_template)
        self.shared_files.append("global_ex_calculator.py")
        self.shared_files.append("ind_ex_calculator.py")

        for f in listdir(coor_path):
            self.shared_files.append(f)

        #-----------------------------------------------------------------------

        parm_url = 'file://%s' % (parm_path)
        self.shared_urls.append(parm_url)

        inp_url = 'file://%s' % (inp_path)
        self.shared_urls.append(inp_url)

        build_inp_url = 'file://%s' % (build_inp_path)
        self.shared_urls.append(build_inp_url)

        global_calc_url = 'file://%s' % (global_calc_path)
        self.shared_urls.append(global_calc_url)

        rstr_template_url = 'file://%s' % (rstr_template_path)
        self.shared_urls.append(rstr_template_url)

        global_calc_url_s = 'file://%s' % (global_calc_path_s)
        self.shared_urls.append(global_calc_url_s)

        ind_calc_url = 'file://%s' % (ind_calc_path)
        self.shared_urls.append(ind_calc_url)

        for f in listdir(coor_path):
            cf_path = join(coor_path,f)
            coor_url = 'file://%s' % (cf_path)
            self.shared_urls.append(coor_url)

    #---------------------------------------------------------------------------
    # 
    def prepare_replica_for_md(self, replica, sd_shared_list):
 
        basename = self.inp_basename
            
        new_input_file = "%s_%d_%d.mdin" % (basename, replica.id, replica.cycle)
        outputname = "%s_%d_%d.mdout" % (basename, replica.id, replica.cycle)
        old_name = "%s_%d_%d" % (basename, replica.id, (replica.cycle-1))

        replica.new_coor = "%s_%d_%d.rst" % (basename, replica.id, replica.cycle)
        replica.new_traj = "%s_%d_%d.mdcrd" % (basename, replica.id, replica.cycle)
        replica.new_info = "%s_%d_%d.mdinfo" % (basename, replica.id, replica.cycle)

        replica.old_coor = old_name + ".rst"

        if (replica.cycle == 0):
            first_step = 0
        elif (replica.cycle == 1):
            first_step = int(self.cycle_steps)
        else:
            first_step = (replica.cycle - 1) * int(self.cycle_steps)

        replica.cycle += 1

        #-----------------------------------------------------------------------  

        input_file = "%s_%d_%d.mdin" % (self.inp_basename, replica.id, (replica.cycle-1))
        output_file = "%s_%d_%d.mdout" % (self.inp_basename, replica.id, (replica.cycle-1))

        stage_out = []
        stage_in = []

        new_coor = replica.new_coor
        new_traj = replica.new_traj
        new_info = replica.new_info
        old_coor = replica.old_coor
        rid = replica.id

        replica_path = "replica_%d/" % (rid)

        if self.down_mdinfo == True:
            info_local = {
                'source':   new_info,
                'target':   new_info,
                'action':   radical.pilot.TRANSFER
            }
            stage_out.append(info_local)

        if self.down_mdout == True:
            output_local = {
                'source':   output_file,
                'target':   output_file,
                'action':   radical.pilot.TRANSFER
            }
            stage_out.append(output_local)

        new_coor_out = {
            'source': new_coor,
            'target': 'staging:///%s' % (replica_path + new_coor),
            'action': radical.pilot.COPY
        }
        stage_out.append(new_coor_out)

        #-----------------------------------------------------------------------
         
        info_out = {
                'source': new_info,
                'target': 'staging:///%s' % (replica_path + new_info),
                'action': radical.pilot.COPY
            }
        stage_out.append(info_out)

        data = {
            "cycle_steps": str(self.cycle_steps),
            "new_restraints" : str(replica.new_restraints),
            "new_temperature" : str(self.init_temperature),
            "amber_input" : str(self.amber_input),
            "new_input_file" : str(new_input_file),
            "us_template": str(self.us_template),
            "cycle" : str(replica.cycle),
            "rstr_val_1" : str(replica.rstr_val_1)
                }
        dump_pre_data = json.dumps(data)

        json_pre_data_bash = dump_pre_data.replace("\\", "")
        json_pre_data_sh = dump_pre_data.replace("\"", "\\\\\"")

        #-----------------------------------------------------------------------
        # umbrella sampling
        data = {
            "replica_id": str(rid),
            "replica_cycle" : str(replica.cycle-1),
            "replicas" : str(self.replicas),
            "base_name" : str(basename),
            "init_temp" : str(replica.new_temperature),
            "amber_input" : str(self.amber_input),
            "amber_parameters": str(self.amber_parameters),
            "new_restraints" : str(replica.new_restraints)
        }
        dump_data = json.dumps(data)
        json_post_data_us = dump_data.replace("\\", "")

        #-----------------------------------------------------------------------

        cu = radical.pilot.ComputeUnitDescription()

        if KERNELS[self.resource]["shell"] == "bourne":
            pre_exec_str = "python input_file_builder.py " + "\'" + json_pre_data_sh + "\'"
            cu.executable = '/bin/sh'
        elif KERNELS[self.resource]["shell"] == "bash":
            pre_exec_str = "python input_file_builder.py " + "\'" + json_pre_data_bash + "\'"
            cu.executable = '/bin/bash'
       
        if replica.cycle == 1:

            amber_str = self.amber_path
            argument_str = " -O " + " -i " + new_input_file + \
                           " -o " + output_file + \
                           " -p " +  self.amber_parameters + \
                           " -c " + replica.coor_file + \
                           " -r " + new_coor + \
                           " -x " + new_traj + \
                           " -inf " + new_info  

            restraints_out = replica.new_restraints
            restraints_out_st = {
                'source': (replica.new_restraints),
                'target': 'staging:///%s' % (replica.new_restraints),
                'action': radical.pilot.COPY
            }
            stage_out.append(restraints_out_st)

            #-------------------------------------------------------------------
            # files needed to be staged in replica dir
            for i in range(3):
                stage_in.append(sd_shared_list[i])

            # us template
            stage_in.append(sd_shared_list[4])

            #-------------------------------------------------------------------            
            # replica coor
            repl_coor = replica.coor_file

            if self.same_coordinates == False:
                while (repl_coor not in self.shared_files):
                    repl_coor = self.c_prefix + "_" + str(replica.indx1-1) + "." + self.c_infix + "_" + str(replica.indx2-1) + "." + self.c_postfix
            else:
                repl_coor = self.c_prefix + "_0." + self.c_infix + "_0." + self.c_postfix

            # index of replica_coor
            c_index = self.shared_files.index(repl_coor) 
            stage_in.append(sd_shared_list[c_index])

            #-------------------------------------------------------------------
            cu.pre_exec = self.pre_exec
            cu.arguments = ['-c', pre_exec_str + "; " + amber_str + argument_str]  
            cu.cores = self.replica_cores
            cu.input_staging = stage_in
            cu.output_staging = stage_out
            cu.mpi = self.replica_mpi
            
        else:
            amber_str = self.amber_path
            argument_str = " -O " + " -i " + new_input_file + " -o " + output_file + \
                           " -p " +  self.amber_parameters + " -c " + old_coor + \
                           " -r " + new_coor + " -x " + new_traj + " -inf " + new_info
            
            # parameters file
            stage_in.append(sd_shared_list[0])

            # base input file ala10_us.mdin
            stage_in.append(sd_shared_list[1])

            # input_file_builder.py
            stage_in.append(sd_shared_list[2])
            #-------------------------------------------------------------------
            # restraint file
            restraints_in_st = {'source': 'staging:///%s' % replica.new_restraints,
                                'target': replica.new_restraints,
                                'action': radical.pilot.COPY
            }
            stage_in.append(restraints_in_st)

            old_coor_st = {'source': 'staging:///%s' % (replica_path + old_coor),
                           'target': (old_coor),
                           'action': radical.pilot.COPY
            }
            stage_in.append(old_coor_st)

            new_coor_out = {
                    'source': new_coor,
                    'target': 'staging:///%s' % (replica_path + new_coor),
                    'action': radical.pilot.COPY
                }
            stage_out.append(new_coor_out)

            cu.pre_exec = self.pre_exec
            cu.arguments = ['-c', pre_exec_str + "; " + amber_str + argument_str] 
            cu.input_staging = stage_in
            cu.output_staging = stage_out
            cu.cores = self.replica_cores
            cu.mpi = self.replica_mpi

        return cu

    #---------------------------------------------------------------------------
    #
    def prepare_replica_for_exchange(self, replicas, replica, sd_shared_list):
           
        basename  = self.inp_basename
        matrix_col = "matrix_column_%s_%s.dat" % (str(replica.id), str(replica.cycle-1))

        stage_out = []
        col_out = {
            'source': matrix_col,
            'target': 'staging:///%s' % matrix_col,
            'action': radical.pilot.COPY
        }
        stage_out.append(col_out)

        all_restraints = []
        for r in replicas:
            all_restraints.append(str(r.new_restraints))

        data = {
            "replica_id": str(replica.id),
            "replica_cycle" : str(replica.cycle-1),
            "replicas" : str(self.replicas),
            "base_name" : str(basename),
            "init_temp" : str(self.init_temperature),
            "new_restraints" : str(replica.new_restraints),
            "amber_path" : str(self.amber_path),
            "amber_input" : str(self.amber_input),
            "all_restraints" : all_restraints
        }
        dump_data = json.dumps(data)
        json_data = dump_data.replace("\\", "")

        """
        rid = replica.id
        data = {
            "replica_id": str(rid),
            "replica_cycle" : str(replica.cycle-1),
            "replicas" : str(self.replicas),
            "base_name" : str(basename),
            "init_temp" : str(replica.new_temperature),
            "new_restraints" : str(replica.new_restraints)
        }
        dump_data = json.dumps(data)
        json_post_data_us_bash = dump_data.replace("\\", "")
        json_post_data_us_sh   = dump_data.replace("\"", "\\\\\"")
        """

        cu = radical.pilot.ComputeUnitDescription()
        cu.executable = "python"
        cu.input_staging  = sd_shared_list[6]
        cu.output_staging = stage_out
        #cu.arguments = ["ind_ex_calculator.py", replica.id, (replica.cycle-1), self.replicas, basename, replica.new_temperature, replica.new_restraints]
        cu.arguments = ["ind_ex_calculator.py", json_data]
        cu.cores = 1
        cu.mpi = False
        
        return cu

    #---------------------------------------------------------------------------
    #
    def prepare_global_ex_calc(self, GL, current_cycle, replicas, sd_shared_list):

        stage_out = []
        stage_in = []

        cycle = replicas[0].cycle-1

        outfile = "pairs_for_exchange_{cycle}.dat".format(cycle=cycle)
        stage_out.append(outfile)

        all_restraints = {}
        all_temperatures = {}
        for repl in replicas:
            all_restraints[str(repl.id)] = str(repl.new_restraints)
            all_temperatures[str(repl.id)] = str(repl.new_temperature)

        data = {
            "current_cycle" : str(cycle),
            "replicas" : str(self.replicas),
            "base_name" : str(self.inp_basename),
            "all_temperatures" : all_temperatures,
            "all_restraints" : all_restraints
        }
        dump_data = json.dumps(data)
        json_data_us = dump_data.replace("\\", "")

        if self.exchange_mpi == True:
            # global_ex_calculator_mpi.py file
            stage_in.append(sd_shared_list[3])

            cu = radical.pilot.ComputeUnitDescription()
            cu.pre_exec = self.pre_exec
            cu.executable = "python"
            cu.input_staging  = stage_in
            cu.arguments = ["global_ex_calculator_mpi.py", json_data_us]

            if self.replicas > 999:
                self.cores = self.replicas / 2
            elif self.cores < self.replicas:
                if (self.replicas % self.cores) != 0:
                    self.logger.info("Number of replicas must be divisible by the number of Pilot cores!")
                    self.logger.info("pilot cores: {0}; replicas {1}".format(self.cores, self.replicas))
                    sys.exit()
                else:
                    cu.cores = self.cores
            elif self.cores >= self.replicas:
                cu.cores = self.replicas
            else:
                self.logger.info("Number of replicas must be divisible by the number of Pilot cores!")
                self.logger.info("pilot cores: {0}; replicas {1}".format(self.cores, self.replicas))
                sys.exit()
            cu.mpi = True         
            cu.output_staging = stage_out
        else:
            # global_ex_calculator.py file
            stage_in.append(sd_shared_list[5])

            cu = radical.pilot.ComputeUnitDescription()
            cu.pre_exec = self.pre_exec
            cu.executable = "python"
            cu.input_staging  = stage_in
            cu.arguments = ["global_ex_calculator.py", str(cycle), str(self.replicas), str(self.inp_basename)]
            cu.cores = 1
            cu.mpi = False         
            cu.output_staging = stage_out

        return cu

    #---------------------------------------------------------------------------
    #
    def perform_swap(self, replica_i, replica_j):
        """Performs an exchange of temperatures

        Arguments:
        replica_i - a replica object
        replica_j - a replica object
        """

        rstr = replica_j.new_restraints
        replica_j.new_restraints = replica_i.new_restraints
        replica_i.new_restraints = rstr

        val = replica_j.rstr_val_1
        replica_j.rstr_val_1 = replica_i.rstr_val_1
        replica_i.rstr_val_1 = val
        # record that swap was performed
        replica_i.swap = 1
        replica_j.swap = 1

    #---------------------------------------------------------------------------
    #
    def do_exchange(self, current_cycle, replicas):
        """
        """

        r1 = None
        r2 = None

        cycle = replicas[0].cycle-1

        infile = "pairs_for_exchange_{cycle}.dat".format(cycle=cycle)
        try:
            f = open(infile)
            lines = f.readlines()
            f.close()
            for l in lines:
                pair = l.split()
                r1_id = int(pair[0])
                r2_id = int(pair[1])
                for r in replicas:
                    if r.id == r1_id:
                        r1 = r
                    if r.id == r2_id:
                        r2 = r
                #---------------------------------------------------------------
                # guard
                if r1 == None:
                    rid = random.randint(0,(len(replicas)-1))
                    r1 = replicas[rid]
                if r2 == None:
                    rid = random.randint(0,(len(replicas)-1))
                    r2 = replicas[rid]
                #---------------------------------------------------------------

                # swap parameters
                self.perform_swap(r1, r2)
                r1.swap = 1
                r2.swap = 1
        except:
            raise    





__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import json
import math
import time
import shutil

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    """
    """

    json_data = sys.argv[1]
    data=json.loads(json_data)

    cycle_steps     = data["cycle_steps"]
    new_salt_concentration = data["new_salt_concentration"]
    amber_input     = data["amber_input"]
    new_input_file  = data["new_input_file"]
    replica_cycle   = int(data["cycle"])

    #---------------------------------------------------------------------------
    # this is for every cycle
    try:
        r_file = open(amber_input, "r")
    except IOError:
        print "Warning: unable to access template file: {0}".format(amber_input) 

    tbuffer = r_file.read()
    r_file.close()

    tbuffer = tbuffer.replace("@nstlim@",cycle_steps)
    #tbuffer = tbuffer.replace("@rstr@",new_restraints)
    #tbuffer = tbuffer.replace("@temp@",new_temperature)
    tbuffer = tbuffer.replace("@salt@",new_salt_concentration)

    try:
        w_file = open(new_input_file, "w")
        w_file.write(tbuffer)
        w_file.close()
    except IOError:
        print "Warning: unable to access file: {0}".format(new_input_file)

    


"""
.. module:: radical.repex.pilot_kernels.pilot_kernel_pattern_a_multi_d
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import time
import math
import json
import datetime
from os import path
import radical.pilot
from pilot_kernels.pilot_kernel import *
import radical.utils.logger as rul

#-------------------------------------------------------------------------------

class PilotKernelPatternSmultiD(PilotKernel):
   
    def __init__(self, inp_file, rconfig):
        """Constructor.

        Arguments:
        inp_file - json input file with Pilot and NAMD related parameters as 
        specified by user 
        """
        PilotKernel.__init__(self, inp_file, rconfig)

        self.name = 'exec-pattern-A-multiD'
        self.logger  = rul.getLogger ('radical.repex', self.name)

        self.sd_shared_list = []

#-------------------------------------------------------------------------------

    def run_simulation(self, replicas, pilot_object, session,  md_kernel):
        """This function runs the main loop of RE simulation for RE pattern B.

        Arguments:
        replicas - list of Replica objects
        pilot_object - radical.pilot.ComputePilot object
        session - radical.pilot.session object, the *root* object for all other 
        RADICAL-Pilot objects 
        md_kernel - an instance of NamdKernelScheme2a class
        """

        # ----------------------------------------------------------------------
        #
        def unit_state_change_cb(unit, state):
            """This is a callback function. It gets called very time a 
            ComputeUnit changes its state.
            """

            if unit:            
                self.logger.info("ComputeUnit '{0:s}' state changed to {1:s}.".format(unit.uid, state) )

                if state == radical.pilot.states.FAILED:
                    self.logger.error("Log: {0:s}".format( unit.as_dict() ) )
                    # restarting the replica
                    #self.logger.info("ComputeUnit '{0:s}' state changed to {1:s}.".format(unit.uid, state) )
                    #unit_manager.submit_units( unit.description )

        # ----------------------------------------------------------------------
        cycles = md_kernel.nr_cycles + 1
                
        unit_manager = radical.pilot.UnitManager(session, scheduler=radical.pilot.SCHED_ROUND_ROBIN)
        unit_manager.register_callback(unit_state_change_cb)
        unit_manager.add_pilots(pilot_object)

        stagein_start = datetime.datetime.utcnow()

        # staging shared input data in
        md_kernel.prepare_shared_data(replicas)

        shared_input_file_urls = md_kernel.shared_urls
        shared_input_files = md_kernel.shared_files

        for i in range(len(shared_input_files)):

            sd_pilot = {'source': shared_input_file_urls[i],
                        'target': 'staging:///%s' % shared_input_files[i],
                        'action': radical.pilot.TRANSFER
            }

            pilot_object.stage_in(sd_pilot)

            sd_shared = {'source': 'staging:///%s' % shared_input_files[i],
                         'target': shared_input_files[i],
                         'action': radical.pilot.COPY
            }
            self.sd_shared_list.append(sd_shared)

        # for performance data collection
        hl_performance_data = {}
        cu_performance_data = {}

        md_kernel.init_matrices(replicas)

        stagein_end = datetime.datetime.utcnow()

        start = datetime.datetime.utcnow()
        #------------------------
        # GL = 0: submit global calculator before
        # GL = 1: submit global calculator after
        GL = 1
        # BULK = 0: do sequential submission
        # BULK = 1: do BULK submission
        BULK = 1
        DIM = 0
        dimensions = md_kernel.dims
        for c in range(0,cycles*dimensions):

            if DIM < dimensions:
                DIM = DIM + 1
            else:
                DIM = 1

            current_cycle = c / dimensions

            if DIM == 1:
                cu_performance_data["cycle_{0}".format(current_cycle)] = {}
                hl_performance_data["cycle_{0}".format(current_cycle)] = {}

            self.logger.info("Performing cycle: {0}".format(current_cycle) )
            
            cu_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)] = {}
            hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)] = {}

            self.logger.info("Dim {0}: preparing {1} replicas for MD run; cycle {2}".format(DIM, md_kernel.replicas, current_cycle) )
            
            submitted_replicas = []
            exchange_replicas = []
            
            #-------------------------------------------------------------------
            # sequential submission
            if BULK == 0:

                #---------------------------------------------------------------
                # submitting unit which determines exchanges between replicas
                if GL == 0:
                    ex_calculator = md_kernel.prepare_global_ex_calc(GL, current_cycle, DIM, replicas, self.sd_shared_list)
                    global_ex_cu = unit_manager.submit_units(ex_calculator)
                #---------------------------------------------------------------

                t1 = datetime.datetime.utcnow()
                for replica in replicas:
                    compute_replica = md_kernel.prepare_replica_for_md(DIM, replicas, replica, self.sd_shared_list)
                    sub_replica = unit_manager.submit_units(compute_replica)
                    submitted_replicas.append(sub_replica)
                t2 = datetime.datetime.utcnow()

                hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_prep"] = {}
                hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_prep"] = (t2-t1).total_seconds()

                if (DIM == 2) and (md_kernel.d2 == 'salt_concentration'):
                    t1 = datetime.datetime.utcnow()
                    unit_manager.wait_units()
                    t2 = datetime.datetime.utcnow()

                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_run"] = {}
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_run"] = (t2-t1).total_seconds()

                    t1 = datetime.datetime.utcnow()
                    for replica in replicas:
                        ex_replica = md_kernel.prepare_replica_for_exchange(DIM, replicas, replica, self.sd_shared_list)
                        sub_replica = unit_manager.submit_units(ex_replica)
                        exchange_replicas.append(sub_replica)
                    t2 = datetime.datetime.utcnow()
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["ex_prep_salt"] = {}
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["ex_prep_salt"] = (t2-t1).total_seconds()
                    
                    #-----------------------------------------------------------
                    # submitting unit which determines exchanges between replicas
                    if GL == 1:
                        ex_calculator = md_kernel.prepare_global_ex_calc(GL, current_cycle, DIM, replicas, self.sd_shared_list)
                        global_ex_cu = unit_manager.submit_units(ex_calculator)
                    #-----------------------------------------------------------

                    t1 = datetime.datetime.utcnow()
                    unit_manager.wait_units()
                    t2 = datetime.datetime.utcnow()

                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["ex_run_salt"] = {}
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["ex_run_salt"] = (t2-t1).total_seconds()
                
                if (DIM != 2) or (md_kernel.d2 != 'salt_concentration'):

                    #-----------------------------------------------------------
                    # submitting unit which determines exchanges between replicas
                    if GL == 1:
                        ex_calculator = md_kernel.prepare_global_ex_calc(GL, current_cycle, DIM, replicas, self.sd_shared_list)
                        global_ex_cu = unit_manager.submit_units(ex_calculator)
                    #-----------------------------------------------------------

                    t1 = datetime.datetime.utcnow()
                    unit_manager.wait_units()
                    t2 = datetime.datetime.utcnow()

                    # md run includes timings for global calc since there isn't
                    # a separate set of CU's for exchange!
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_run"] = {}
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_run"] = (t2-t1).total_seconds()
                else:
                    unit_manager.wait_units()
            #-------------------------------------------------------------------
            # BULK submision
            else:

                #---------------------------------------------------------------
                # submitting unit which determines exchanges between replicas
                if GL == 0:
                    ex_calculator = md_kernel.prepare_global_ex_calc(GL, current_cycle, DIM, replicas, self.sd_shared_list)
                    global_ex_cu = unit_manager.submit_units(ex_calculator)
                #---------------------------------------------------------------
                
                c_replicas = []
                t1 = datetime.datetime.utcnow()
                for replica in replicas:
                    # DIM, replicas, replica, self.sd_shared_list
                    compute_replica = md_kernel.prepare_replica_for_md(DIM, replicas, replica, self.sd_shared_list)
                    c_replicas.append(compute_replica)
                t2 = datetime.datetime.utcnow()
                submitted_replicas = unit_manager.submit_units(c_replicas)

                hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_prep"] = {}
                hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_prep"] = (t2-t1).total_seconds()

                #---------------------------------------------------------------
                
                e_replicas = []
                if (DIM == 2) and (md_kernel.d2 == 'salt_concentration'):
                    t1 = datetime.datetime.utcnow()
                    unit_manager.wait_units()
                    t2 = datetime.datetime.utcnow()

                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_run"] = {}
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_run"] = (t2-t1).total_seconds()

                    t1 = datetime.datetime.utcnow()
                    for replica in replicas:
                        ex_replica = md_kernel.prepare_replica_for_exchange(DIM, replicas, replica, self.sd_shared_list)
                        e_replicas.append(ex_replica)
                    exchange_replicas = unit_manager.submit_units(e_replicas) 
                    t2 = datetime.datetime.utcnow()

                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["ex_prep_salt"] = {}
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["ex_prep_salt"] = (t2-t1).total_seconds()

                    #-----------------------------------------------------------
                    # submitting unit which determines exchanges between replicas
                    if GL == 1:
                        ex_calculator = md_kernel.prepare_global_ex_calc(GL, current_cycle, DIM, replicas, self.sd_shared_list)
                        global_ex_cu = unit_manager.submit_units(ex_calculator)
                    #-----------------------------------------------------------
                    
                    t1 = datetime.datetime.utcnow()
                    unit_manager.wait_units()
                    t2 = datetime.datetime.utcnow()

                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["ex_run_salt"] = {}
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["ex_run_salt"] = (t2-t1).total_seconds()
                
                
                if (DIM != 2) or (md_kernel.d2 != 'salt_concentration'):
                    #-----------------------------------------------------------
                    # submitting unit which determines exchanges between replicas
                    if GL == 1:
                        ex_calculator = md_kernel.prepare_global_ex_calc(GL, current_cycle, DIM, replicas, self.sd_shared_list)
                        global_ex_cu = unit_manager.submit_units(ex_calculator)
                    #-----------------------------------------------------------

                    t1 = datetime.datetime.utcnow()
                    unit_manager.wait_units()
                    t2 = datetime.datetime.utcnow()

                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_run"] = {}
                    hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["md_run"] = (t2-t1).total_seconds()
                else:
                    unit_manager.wait_units()
            #-------------------------------------------------------------------

            cu_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["run_{0}".format("MD")] = {}
            for cu in submitted_replicas:
                cu_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["run_{0}".format("MD")]["cu.uid_{0}".format(cu.uid)] = cu
            
            cu_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["run_{0}".format("EX_salt")] = {}
            for cu in exchange_replicas:
                cu_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["run_{0}".format("EX_salt")]["cu.uid_{0}".format(cu.uid)] = cu

            cu_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["run_{0}".format("GLOBAL_EX")] = {}
            cu_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["run_{0}".format("GLOBAL_EX")]["cu.uid_{0}".format(global_ex_cu.uid)] = global_ex_cu

            #-------------------------------------------------------------------
            #               
            t1 = datetime.datetime.utcnow()
            for r in submitted_replicas:
                if r.state != radical.pilot.DONE:
                    self.logger.error('ERROR: In D%d MD-step failed for unit:  %s' % (DIM, r.uid))

            if len(exchange_replicas) > 0:
                for r in exchange_replicas:
                    if r.state != radical.pilot.DONE:
                        self.logger.error('ERROR: In D%d Exchange-step failed for unit:  %s' % (DIM, r.uid))

            if global_ex_cu.state != radical.pilot.DONE:
                self.logger.error('ERROR: In D%d Global-Exchange-step failed for unit:  %s' % (DIM, global_ex_cu.uid))

            # do exchange of parameters                     
            md_kernel.do_exchange(current_cycle, DIM, replicas)
            t2 = datetime.datetime.utcnow()
                
            hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["POST_PROC"] = {}
            hl_performance_data["cycle_{0}".format(current_cycle)]["dim_{0}".format(DIM)]["POST_PROC"] = (t2-t1).total_seconds()
            
            #-------------------------------------------------------------------
            # performance data
            outfile = "execution_profile_{mysession}.csv".format(mysession=session.uid)
            with open(outfile, 'a') as f:
                
                #---------------------------------------------------------------
                #
                head = "Cycle; Dim; Run; Duration"
                f.write("{row}\n".format(row=head))

                hl_cycle = "cycle_{0}".format(current_cycle)
                hl_dim   = "dim_{0}".format(DIM)
                
                for run in hl_performance_data[hl_cycle][hl_dim].keys():
                    dur = hl_performance_data[hl_cycle][hl_dim][run]

                    row = "{Cycle}; {Dim}; {Run}; {Duration}".format(
                        Duration=dur,
                        Cycle=hl_cycle,
                        Dim=hl_dim,
                        Run=run)

                    f.write("{r}\n".format(r=row))

                #---------------------------------------------------------------
                head = "CU_ID; Scheduling; StagingInput; PendingAgentInputStaging; AgentStagingInput; Allocating; Executing; PendingAgentOutputStaging; PendingOutputStaging; StagingOutput; Done; Cycle; Dim; Run;"
                f.write("{row}\n".format(row=head))
               
                for run in cu_performance_data[hl_cycle][hl_dim].keys():
                    for cid in cu_performance_data[hl_cycle][hl_dim][run].keys():
                        cu = cu_performance_data[hl_cycle][hl_dim][run][cid]
                        st_data = {}
                        for st in cu.state_history:
                            st_dict = st.as_dict()
                            st_data["{0}".format( st_dict["state"] )] = {}
                            st_data["{0}".format( st_dict["state"] )] = st_dict["timestamp"]
                       
                        row = "{uid}; {Scheduling}; {StagingInput}; {PendingAgentInputStaging}; {AgentStagingInput}; {Allocating}; {Executing}; {PendingAgentOutputStaging}; {PendingOutputStaging}; {StagingOutput}; {Done}; {Cycle}; {Dim}; {Run}".format(
                            uid=cu.uid,
                            #Unscheduled=st_data['Unscheduled'],
                            Scheduling=st_data['Scheduling'],
                            #PendingInputStaging = st_data['PendingInputStaging'],
                            StagingInput=st_data['StagingInput'],
                            PendingAgentInputStaging=st_data['PendingAgentInputStaging'],
                            AgentStagingInput=st_data['AgentStagingInput'],
                            #PendingExecution=st_data['PendingExecution'],
                            Allocating=st_data['Allocating'],
                            Executing=st_data['Executing'],
                            PendingAgentOutputStaging=st_data['PendingAgentOutputStaging'],
                            #AgentStagingOutput=st_data['AgentStagingOutput'],
                            PendingOutputStaging=st_data['PendingOutputStaging'],
                            StagingOutput=st_data['StagingOutput'],
                            Done=st_data['Done'],
                            Cycle=hl_cycle,
                            Dim=hl_dim,
                            Run=run)

                        f.write("{r}\n".format(r=row))
            
        #-----------------------------------------------------------------------
        # end of loop

        outfile = "execution_profile_{mysession}.csv".format(mysession=session.uid)
        with open(outfile, 'a') as f:
            # RAW SIMULATION TIME
            end = datetime.datetime.utcnow()
            stagein_time = (stagein_end-stagein_start).total_seconds()
            raw_simulation_time = (end-start).total_seconds()
            f.write("Total simulaiton time: {row}\n".format(row=raw_simulation_time))
            f.write("Stage-in time: {row}\n".format(row=stagein_time))




"""
.. module:: radical.repex.md_kernels.amber_kernels_tex.matrix_calculator_tex
.. moduleauthor::  <antons.treikalis@rutgers.edu>
.. moduleauthor::  <haoyuan.chen@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import json
import math
import time
import shutil

#-------------------------------------------------------------------------------
#
def reduced_energy(temperature, potential):
    """Calculates reduced energy.

    Arguments:
    temperature - replica temperature
    potential - replica potential energy

    Returns:
    reduced enery of replica
    """
    kb = 0.0019872041    #boltzmann const in kcal/mol
    if temperature != 0:
        beta = 1. / (kb*temperature)
    else:
        beta = 1. / kb     
    return float(beta * potential)

#-------------------------------------------------------------------------------
#
def get_historical_data(replica_path, history_name):
    """Retrieves temperature and potential energy from simulation output file 
    .history file. This file is generated after each simulation run. The function 
    searches for directory where .history file recides by checking all 
    computeUnit directories on target resource.

    Arguments:
    history_name - name of .history file for a given replica. 

    Returns:
    data[0] - temperature obtained from .history file
    data[1] - potential energy obtained from .history file
    path_to_replica_folder - path to computeUnit directory on a target resource 
    where all input/output files for a given replica recide.
       Get temperature and potential energy from mdinfo file.
    """

    home_dir = os.getcwd()
    if replica_path != None:
        path = "../staging_area" + replica_path
        try:
            os.chdir(path)
        except:
            raise

    temp = 0.0    #temperature
    eptot = 0.0   #potential

    try:
        f = open(history_name)
        lines = f.readlines()
        f.close()
        path_to_replica_folder = os.getcwd()
        for i in range(len(lines)):
            if "TEMP(K)" in lines[i]:
                temp = float(lines[i].split()[8])
            elif "EPtot" in lines[i]:
                eptot = float(lines[i].split()[8])
    except:
        os.chdir(home_dir)
        raise 

    os.chdir(home_dir)
    
    return temp, eptot, path_to_replica_folder

#-------------------------------------------------------------------------------
#
if __name__ == '__main__':

    """
    json_data = sys.argv[1]
    data=json.loads(json_data)

    replica_id = data["replica_id"]
    replica_cycle = data["replica_cycle"]
    base_name = data["base_name"]
    replicas = int(data["replicas"])
    amber_parameters = data["amber_parameters"]
    init_temp = data["init_temp"]


    temp_group = data["current_group"]
    current_group = []
    """

    argument_list = str(sys.argv)
    replica_id = str(sys.argv[1])
    replica_cycle = str(sys.argv[2])
    replicas = int(str(sys.argv[3]))
    base_name = str(sys.argv[4])
    init_temp = str(sys.argv[5])

    pwd = os.getcwd()


    #for i in temp_group:
    #    current_group.append(int(i))
     
    # getting history data for self
    history_name = base_name + "_" + replica_id + "_" + replica_cycle + ".mdinfo"

    # init swap column
    swap_column = [0.0]*replicas

    #---------------------------------------------------------------------------
    success = 0
    attempts = 0
    while (success == 0):
        try:
            replica_temp, replica_energy, path_to_replica_folder = get_historical_data("/", history_name)
            print "Got history data for self!"
            success = 1
        except:
            print "Waiting for self (history file)"
            time.sleep(1)
            attempts += 1
            if attempts >= 5:
                #---------------------------------------------------------------
                # writing to file
                try:
                    outfile = "matrix_column_{replica}_{cycle}.dat".format(cycle=replica_cycle, replica=replica_id )
                    with open(outfile, 'w+') as f:
                        row_str = ""
                        for item in swap_column:
                            if len(row_str) != 0:
                                row_str = row_str + " " + str(item)
                            else:
                                row_str = str(item)
                            f.write(row_str)
                            f.write('\n')
                            row_str = str(replica_id) + " " + str(replica_cycle) + " " + str(init_temp)
                            f.write(row_str)
                        f.close()
                except IOError:
                    print 'Error: unable to create column file %s for replica %s' % (outfile, replica_id)
                #---------------------------------------------------------------
                sys.exit("Amber run failed, matrix_swap_column_x_x.dat populated with zeros")
            pass

    # getting history data for all replicas
    temperatures = [0.0]*replicas
    energies = [0.0]*replicas

    # for self
    temperatures[int(replica_id)] = replica_temp
    energies[int(replica_id)] = replica_energy

    for j in range(replicas):
        # for self already processed
        if j != int(replica_id):
            success = 0
            attempts = 0
            history_name = base_name + "_" + str(j) + "_" + replica_cycle + ".mdinfo" 
            while (success == 0):
                try:
                    rj_temp, rj_energy, temp = get_historical_data("/", history_name)
                    temperatures[j] = rj_temp
                    energies[j] = rj_energy

                    success = 1
                    print "Success processing replica: %s" % j
                except:
                    print "Waiting for replica: %s" % j
                    time.sleep(1)
                    attempts += 1
                    # some of the replicas failed
                    # set temperature and energy for those replicas as -1.0
                    if attempts >= 5:
                        temperatures[j] = -1.0
                        energies[j] = -1.0
                        success = 1
                        print "Replica %d failed, initialized temperatures[j] and energies[j] to -1.0" % j
                    pass

    print "got history data for other replicas!"

    for j in range(replicas):     
        swap_column[j] = reduced_energy(temperatures[j], replica_energy)

    #---------------------------------------------------------------------------
    # writing to file
    
    try:
        outfile = "matrix_column_{replica}_{cycle}.dat".format(cycle=replica_cycle, replica=replica_id )
        with open(outfile, 'w+') as f:
            row_str = ""
            for item in swap_column:
                if len(row_str) != 0:
                    row_str = row_str + " " + str(item)
                else:
                    row_str = str(item)
            f.write(row_str)
            f.write('\n')
            row_str = replica_id + " " + replica_cycle + " " + init_temp
            f.write(row_str)
        f.close()

    except IOError:
        print 'Error: unable to create column file %s for replica %s' % (outfile, replica_id)

    


"""
.. module:: radical.repex.replica_cleanup
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"


import os
import sys
import shutil

#-----------------------------------------------------------------------------------------------------------------------------------

def move_output_files(work_dir_local, inp_basename, replicas):
    """Moves all files starting with <inp_basename> to replica directories. These are files generated and 
    transferred to home dorectory as a result of NAMD simulation. This includes .coor, .xcs and other files.
    In addition to that files representing columns of swap matrix (<matrix_column_x.dat>) are transferred as well.

    Arguments:
    replicas - list of Replica objects
    """

    #----------------------------------------------------------
    # moving shared files
    #----------------------------------------------------------
    dir_path = "%s/shared_files" % (work_dir_local)
    if not os.path.exists(dir_path):
        try:
            os.makedirs(dir_path)
        except: 
            raise

    pairs_name = "pairs_for_exchange_"
    files = os.listdir( work_dir_local )

    for item in files:
        if (item.startswith(pairs_name)):
            source =  work_dir_local + "/" + str(item)
            destination = dir_path + "/"
            shutil.move( source, destination)

    #----------------------------------------------------------
    # moving individual files
    #----------------------------------------------------------

    for r in range(len(replicas)):
        dir_path = "%s/replica_%d" % (work_dir_local, r )
        if not os.path.exists(dir_path):
            try:
                os.makedirs(dir_path)
            except: 
                raise

        files = os.listdir( work_dir_local )

        base_name =  inp_basename[:-5] + "_%s_" % replicas[r].id
        col_name = "matrix_column" + "_%s_" % replicas[r].id
        rst_name = "ala10_us.RST." + "%s" % replicas[r].id

        for item in files:
            if (item.startswith(base_name) or item.startswith(col_name) or item.startswith(rst_name)):
                source =  work_dir_local + "/" + str(item)
                destination = dir_path + "/"
                shutil.move( source, destination)

#-----------------------------------------------------------------------------------------------------------------------------------

def clean_up(work_dir_local, replicas):
    """Automates deletion of directories of individual replicas and all files in 
    those directories after the simulation.   

    Arguments:
    replicas - list of Replica objects
    """
    for r in range(len(replicas)):
        dir_path = "%s/replica_%d" % ( work_dir_local, replicas[r].id )
        shutil.rmtree(dir_path)

    dir_path = "%s/shared_files" % ( work_dir_local )
    shutil.rmtree(dir_path)



"""
.. module:: radical.repex.pilot_kernels.pilot_kernel
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import math
import json
import time
import random
import datetime
from os import path
import radical.pilot
from random import randint
from kernels.kernels import KERNELS
import radical.utils.logger as rul

#-------------------------------------------------------------------------------

class PilotKernel(object):
    """
    """
    def __init__(self, inp_file, rconfig):
        """Constructor.

        Arguments:
        inp_file - json input file with Pilot and NAMD related parameters as specified by user 
        """

        self.name = 'exec'
        self.logger  = rul.getLogger ('radical.repex', self.name)
        
        # pilot parameters
        self.resource = rconfig['target']['resource']
        try:
            self.sandbox = rconfig['target']['sandbox']
        except:
            self.sandbox = None
      
        self.user = rconfig['target']['username']
        try:
            self.password = rconfig['target']['password']
        except:
            self.password = None
        try:
            self.project = rconfig['target']['project']
        except:
            self.project = None
 
        try:
            self.queue = rconfig['target']['queue']
        except:
            self.queue = None

        try:
            self.queue = rconfig['target']['queue']
        except:
            self.queue = None
        try:
            self.cores = int(rconfig['target']['cores'])
        except:
            self.logger.info("Field 'cores' must be defined" )
            
        self.runtime = int(rconfig['target']['runtime'])
        try:
            self.dburl = rconfig['target']['mongo_url']
        except:
            self.logger.info("Using default Mongo DB url" )
            self.dburl = "mongodb://ec2-54-221-194-147.compute-1.amazonaws.com:24242/cdi-test"
        cleanup = rconfig['target']['cleanup']
        if (cleanup == "True"):
            self.cleanup = True
        else:
            self.cleanup = False 

    # --------------------------------------------------------------------------
    #
    def launch_pilot(self):
 
        # ----------------------------------------------------------------------
        #
        def pilot_state_cb(pilot, state):
           
            if pilot:
                self.logger.info("ComputePilot '{0}' state changed to {1}.".format(pilot.uid, state) )

                if state == radical.pilot.states.FAILED:
                    self.logger.error("Pilot error: {0}".format(pilot.log) )
                    self.logger.error("RepEx execution FAILED.")
                    # sys.exit(1)
        # ----------------------------------------------------------------------
        
        session = None
        pilot_manager = None
        pilot_object = None
   
        session = radical.pilot.Session(database_url=self.dburl)
        self.logger.info("Session ID: {0}".format(session.uid) )

        try:
            # Add an ssh identity to the session.
            cred = radical.pilot.Context('ssh')
            cred.user_id = self.user
            session.add_context(cred)

            pilot_manager = radical.pilot.PilotManager(session=session)
            pilot_manager.register_callback(pilot_state_cb)

            pilot_description = radical.pilot.ComputePilotDescription()
            if self.resource.startswith("localhost"):
                pilot_description.resource = "local.localhost"
            else:
                pilot_description.resource = self.resource

            if(self.sandbox != None):
                pilot_description.sandbox = str(self.sandbox)

            if(self.project != None):
                pilot_description.project = str(self.project)   

            if(self.queue != None):
                pilot_description.queue = str(self.queue)

            pilot_description.cores = self.cores
            pilot_description.runtime = self.runtime
            pilot_description.cleanup = self.cleanup
            
            """
            pilot_description._config = {'number_of_workers' : {'StageinWorker'   :  8,
                                                                'ExecWorker'      :  8,
                                                                'StageoutWorker'  :  8,
                                                                'UpdateWorker'    :  1},
                                         'blowup_factor'     : {'Agent'           :  1,
                                                                'stagein_queue'   :  1,
                                                                'StageinWorker'   :  1,
                                                                'schedule_queue'  :  1,
                                                                'Scheduler'       :  1,
                                                                'execution_queue' :  1,
                                                                'ExecWorker'      :  1,
                                                                'watch_queue'     :  1,
                                                                'Watcher'         :  1,
                                                                'stageout_queue'  :  1,
                                                                'StageoutWorker'  :  1,
                                                                'update_queue'    :  1,
                                                                'UpdateWorker'    :  1},
                                         'drop_clones'       : {'Agent'           :  1,
                                                                'stagein_queue'   :  1,
                                                                'StageinWorker'   :  1,
                                                                'schedule_queue'  :  1,
                                                                'Scheduler'       :  1,
                                                                'execution_queue' :  1,
                                                                'ExecWorker'      :  1,
                                                                'watch_queue'     :  1,
                                                                'Watcher'         :  1,
                                                                'stageout_queue'  :  1,
                                                                'StageoutWorker'  :  1,
                                                                'update_queue'    :  1,
                                                                'UpdateWorker'    :  1}}
            """
            
            pilot_object = pilot_manager.submit_pilots(pilot_description)
            
            # we wait for the pilot to start running on resource
            self.logger.info("Pilot ID: {0}".format(pilot_object.uid) )
            pilot_manager.wait_pilots(pilot_object.uid,'Active') 

        except radical.pilot.PilotException, ex:
            self.logger.error("Error: {0}".format(ex))

        return pilot_manager, pilot_object, session




"""
.. module:: radical.repex.namd_kernels.namd_kernel_scheme_4
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
from os import path
import radical.pilot
from kernels.kernels import KERNELS
from namd_kernel_tex import *
import namd_kernels_tex.namd_matrix_calculator_scheme_4

#-----------------------------------------------------------------------------------------------------------------------------------

class NamdKernelTexScheme4(NamdKernelTex):
    """This class is responsible for performing all operations related to NAMD for RE scheme 2.
    In this class is determined how replica input files are composed, how exchanges are performed, etc.

    RE scheme 4:


    """
    def __init__(self, inp_file, work_dir_local):
        """Constructor.

        Arguments:
        inp_file - package input file with Pilot and NAMD related parameters as specified by user 
        work_dir_local - directory from which main simulation script was invoked
        """
        NamdKernelTex.__init__(self, inp_file, work_dir_local)

        try:
            self.cycle_time = int(inp_file['input.MD']['cycle_time'])
        except:
            self.cycle_time = 3

#----------------------------------------------------------------------------------------------------------------------------------

    def build_input_file(self, replica):
        """Generates input file for individual replica, based on template input file. Tokens @xxx@ are
        substituted with corresponding parameters. 

        old_output_root @oldname@ determines which .coor .vel and .xsc files are used for next cycle

        Arguments:
        replica - a single Replica object
        """

        basename = self.inp_basename[:-5]
        template = self.inp_basename
            
        new_input_file = "%s_%d_%d.namd" % (basename, replica.id, replica.cycle)
        outputname = "%s_%d_%d" % (basename, replica.id, replica.cycle)
        old_name = "%s_%d_%d" % (basename, replica.id, (replica.cycle-1))

        # these are required to transfer outputs back to laptop, but
        # we don't know at which i_run md was stopped, so in principle
        # we can only transfer during the exchange run using calculator
        #replica.new_coor = outputname + ".coor"
        #replica.new_vel = outputname + ".vel"
        #replica.new_ext_system = outputname + ".xsc"

        # not sure if current use of first time step makes a lot of sense
        # why first step needs to be incremented???
        if (replica.cycle == 0):
            first_step = 0
        elif (replica.cycle == 1):
            first_step = int(self.cycle_steps)
        else:
            first_step = (replica.cycle - 1) * int(self.cycle_steps)

        # if, else below seems OK
        if (replica.cycle == 0):
            old_name = "%s_%d_%d" % (basename, replica.id, (replica.cycle-1)) 
            structure = self.namd_structure
            coordinates = self.namd_coordinates
            parameters = self.namd_parameters
        else:
            ##################################
            # changing old path from absolute 
            # to relative so that NAMD can 
            # process it
            ##################################
            path_list = []
            for char in reversed(replica.old_path):
                if char == '/': break
                path_list.append( char )

            modified_old_path = ''
            for char in reversed( path_list ):
                modified_old_path += char
            modified_old_path = '../' + modified_old_path.rstrip()
            ##################################

            ##################################
            # changing first path from absolute 
            # to relative so that NAMD can 
            # process it
            ##################################
            path_list = []
            for char in reversed(replica.first_path):
                if char == '/': break
                path_list.append( char )

            modified_first_path = ''
            for char in reversed( path_list ):
                modified_first_path += char
            modified_first_path = '../' + modified_first_path.rstrip()
            ##################################

            old_name = modified_old_path + "/%s_%d_%d" % (basename, replica.id, (replica.cycle-1))
            structure = modified_first_path + "/" + self.namd_structure
            coordinates = modified_first_path + "/" + self.namd_coordinates
            parameters = modified_first_path + "/" + self.namd_parameters


        # substituting tokens in main replica input file 
        try:
            r_file = open( (os.path.join((self.work_dir_local + "/namd_inp/"), template)), "r")
        except IOError:
            print 'Warning: unable to access template file %s' % template

        tbuffer = r_file.read()
        r_file.close()

        tbuffer = tbuffer.replace("@swap@",str(replica.swap))
        tbuffer = tbuffer.replace("@ot@",str(replica.old_temperature))
        tbuffer = tbuffer.replace("@nt@",str(replica.new_temperature))
        tbuffer = tbuffer.replace("@steps@",str(self.cycle_steps))
        tbuffer = tbuffer.replace("@rid@",str(replica.id))
        tbuffer = tbuffer.replace("@somename@",str(outputname))                                        #
        tbuffer = tbuffer.replace("@oldname@",str(old_name))
        tbuffer = tbuffer.replace("@cycle@",str(replica.cycle))
        tbuffer = tbuffer.replace("@firststep@",str(first_step))
        tbuffer = tbuffer.replace("@structure@", str(structure))
        tbuffer = tbuffer.replace("@coordinates@", str(coordinates))
        tbuffer = tbuffer.replace("@parameters@", str(parameters))

        tbuffer = tbuffer.replace("@stopped_run@", str(replica.stopped_run))
        
        replica.cycle += 1
        # write out
        try:
            w_file = open( new_input_file, "w")
            w_file.write(tbuffer)
            w_file.close()
        except IOError:
            print 'Warning: unable to access file %s' % new_input_file

#-----------------------------------------------------------------------------------------------------------------------------------

    def prepare_replicas_for_md(self, replicas):
        """Creates a list of ComputeUnitDescription objects for MD simulation step. Here are
        specified input/output files to be transferred to/from target resource. Note: input 
        files for first and subsequent simulaition cycles are different.  

        Arguments:
        replicas - list of Replica objects

        Returns:
        compute_replicas - list of radical.pilot.ComputeUnitDescription objects
        """
        compute_replicas = []
        for r in range(len(replicas)):
            self.build_input_file(replicas[r])
            input_file = "%s_%d_%d.namd" % (self.inp_basename[:-5], replicas[r].id, (replicas[r].cycle-1))

            new_coor = replicas[r].new_coor
            new_vel = replicas[r].new_vel
            new_history = replicas[r].new_history
            new_ext_system = replicas[r].new_ext_system

            old_coor = replicas[r].old_coor
            old_vel = replicas[r].old_vel
            old_ext_system = replicas[r].old_ext_system 

            # only for first cycle we transfer structure, coordinates and parameters files
            if replicas[r].cycle == 1:
                cu = radical.pilot.ComputeUnitDescription()
                cu.pre_exec    = KERNELS[self.resource]["kernels"]["namd"]["pre_execution"]
                cu.executable = self.namd_path
                cu.arguments = [input_file]
                cu.cores = replicas[r].cores
                cu.mpi = False
                structure = self.work_dir_local + "/" + self.inp_folder + "/" + self.namd_structure
                coords = self.work_dir_local + "/" + self.inp_folder + "/" + self.namd_coordinates
                params = self.work_dir_local + "/" + self.inp_folder + "/" + self.namd_parameters
                cu.input_staging = [str(input_file), str(structure), str(coords), str(params)]
 
                #cu.output_data = [new_coor, new_vel, new_history, new_ext_system ]
                compute_replicas.append(cu)
            else:
                cu = radical.pilot.ComputeUnitDescription()
                cu.pre_exec    = KERNELS[self.resource]["kernels"]["namd"]["pre_execution"]
                cu.executable = self.namd_path
                cu.arguments = [input_file]
                cu.cores = replicas[r].cores
                cu.mpi = False
                cu.input_staging = [str(input_file)]
                # in principle it is not required to transfer simulation output files in order to 
                # perform the next cycle; this is done mainly to have these files on local system;
                # an alternative approach would be to transfer all the files at the end of the simulation

                # later we need to transfer these files back to laptop
                #cu.output_data = [new_coor, new_vel, new_history, new_ext_system ]
                compute_replicas.append(cu)

        return compute_replicas

#-----------------------------------------------------------------------------------------------------------------------------------

    def prepare_replicas_for_exchange(self, replicas):
        """Creates a list of ComputeUnitDescription objects for exchange step on resource.
        Number of matrix_calculator_scheme_2.py instances invoked on resource is equal to the number 
        of replicas. 

        Arguments:
        replicas - list of Replica objects

        Returns:
        exchange_replicas - list of radical.pilot.ComputeUnitDescription objects
        """

        exchange_replicas = []
        for r in range(len(replicas)):
           
            # name of the file which contains swap matrix column data for each replica
            matrix_col = "matrix_column_%s_%s.dat" % (r, (replicas[r].cycle-1))
            basename = self.inp_basename[:-5]
            cu = radical.pilot.ComputeUnitDescription()
            cu.executable = "python"
            # each scheme has it's own calculator!
            calculator_path = os.path.dirname(namd_kernels_tex.namd_matrix_calculator_scheme_4.__file__)
            calculator = calculator_path + "/namd_matrix_calculator_scheme_4.py"
            cu.input_staging = [str(calculator)]
            cu.arguments = ["namd_matrix_calculator_scheme_4.py", r, (replicas[r].cycle-1), len(replicas), basename]
            cu.cores = 1            
            cu.output_staging = [str(matrix_col)]
            exchange_replicas.append(cu)

        return exchange_replicas

        


"""
.. module:: radical.repex.md_kernles_tex.amber_kernels_tex.kernel_pattern_s_tex
.. moduleauthor::  <antons.treikalis@rutgers.edu>
.. moduleauthor::  <haoyuan.chen@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import time
import math
import json
import random
import shutil
import datetime
from os import path
import radical.pilot
import radical.utils.logger as rul
from kernels.kernels import KERNELS
import amber_kernels_tex.input_file_builder
#import amber_kernels_tex.global_ex_calculator_mpi
import amber_kernels_tex.global_ex_calculator
import amber_kernels_tex.ind_ex_calculator
from replicas.replica import *

#-------------------------------------------------------------------------------

class KernelPatternStex(object):
    
    def __init__(self, inp_file, rconfig,  work_dir_local):
        """Constructor.

        Arguments:
        inp_file - package input file with Pilot and NAMD related parameters as 
        specified by user 
        work_dir_local - directory from which main simulation script was invoked
        """

        self.name = 'amber-tremd-patternA'
        self.ex_name = 'temperature'
        self.logger  = rul.getLogger ('radical.repex', self.name)

        if 'number_of_cycles' in inp_file['remd.input']:
            self.nr_cycles = int(inp_file['remd.input']['number_of_cycles'])
        else:
            self.nr_cycles = None

        if 'replica_mpi' in inp_file['remd.input']:
            mpi = inp_file['remd.input']['replica_mpi']
            if mpi == "True":
                self.replica_mpi = True
            else:
                self.replica_mpi = False
        else:
            self.replica_mpi = False

        if 'exchange_mpi' in inp_file['remd.input']:
            mpi = inp_file['remd.input']['exchange_mpi']
            if mpi == "True":
                self.exchange_mpi = True
            else:
                self.exchange_mpi = False
        else:
            self.exchange_mpi = False

        if 'replica_cores' in inp_file['remd.input']:
            self.replica_cores = int(inp_file['remd.input']['replica_cores'])
        else:
            self.replica_cores = 1

        self.resource = rconfig['target']['resource']
        self.cores    = int(rconfig['target']['cores'])
        self.pre_exec = KERNELS[self.resource]["kernels"]["amber"]["pre_execution"]

        if 'download_mdinfo' in inp_file['remd.input']:
            if inp_file['remd.input']['download_mdinfo'] == 'True':
                self.down_mdinfo = True
            else:
                self.down_mdinfo = False
        else:
            self.down_mdinfo = True

        if 'download_mdout' in inp_file['remd.input']:
            if inp_file['remd.input']['download_mdout'] == 'True':
                self.down_mdout = True
            else:
                self.down_mdout = False
        else:
            self.down_mdout = True

        if self.resource == 'local.localhost':
            if 'amber_path' in inp_file['remd.input']:
                self.amber_path = inp_file['remd.input']['amber_path']
            else:
                self.logger.info("Amber path for {0} is not defined".format( rconfig['target']['resource'] ) )
                sys.exit(1)
        else:
            if 'amber_path' in inp_file['remd.input']:
                self.amber_path = inp_file['remd.input']['amber_path']
            else:
                self.logger.info("Using default Amber path for: {0}".format( rconfig['target']['resource'] ) )
                try:
                    self.amber_path = KERNELS[self.resource]["kernels"]["amber"]["executable"]
                except:
                    self.logger.info("Amber path for {0} is not defined".format( rconfig['target']['resource'] ) )

        self.input_folder = inp_file['remd.input']['input_folder']   
        self.amber_coordinates = inp_file['remd.input']['amber_coordinates']
        self.amber_parameters = inp_file['remd.input']['amber_parameters']
        self.amber_input = inp_file['remd.input']['amber_input']
        self.input_folder = inp_file['remd.input']['input_folder']
        self.inp_basename = inp_file['remd.input']['input_file_basename']
        self.replicas = int(inp_file['remd.input']['number_of_replicas'])
        self.min_temp = float(inp_file['remd.input']['min_temperature'])
        self.max_temp = float(inp_file['remd.input']['max_temperature'])
        self.cycle_steps = int(inp_file['remd.input']['steps_per_cycle'])
        self.work_dir_local = work_dir_local

        self.current_cycle = -1
 
        self.shared_urls = []
        self.shared_files = []

        self.all_temp_list = []

    #---------------------------------------------------------------------------
    #
    def prepare_shared_data(self):
 
        parm_path = self.work_dir_local + "/" + self.input_folder + "/" + \
                    self.amber_parameters
        coor_path = self.work_dir_local + "/" + self.input_folder + "/" + \
                    self.amber_coordinates
                    
        input_template = self.inp_basename[:-5] + ".mdin"
        input_template_path = self.work_dir_local + "/" + self.input_folder + "/" + input_template

        build_inp = os.path.dirname(amber_kernels_tex.input_file_builder.__file__)
        build_inp_path = build_inp + "/input_file_builder.py"

        global_calc = os.path.dirname(amber_kernels_tex.global_ex_calculator.__file__)
        global_calc_path = global_calc + "/global_ex_calculator.py"

        global_calc_s = os.path.dirname(amber_kernels_tex.global_ex_calculator.__file__)
        global_calc_path_s = global_calc_s + "/global_ex_calculator.py"

        ind_calc = os.path.dirname(amber_kernels_tex.ind_ex_calculator.__file__)
        ind_calc_path = ind_calc + "/ind_ex_calculator.py"

        #-----------------------------------------------------------------------
        self.shared_files.append(self.amber_parameters)
        self.shared_files.append(self.amber_coordinates)
        self.shared_files.append(input_template)
        self.shared_files.append("input_file_builder.py")
        self.shared_files.append("global_ex_calculator.py")
        self.shared_files.append("global_ex_calculator.py")
        self.shared_files.append("ind_ex_calculator.py")

        parm_url = 'file://%s' % (parm_path)
        self.shared_urls.append(parm_url)  

        coor_url = 'file://%s' % (coor_path)
        self.shared_urls.append(coor_url)   

        inp_url = 'file://%s' % (input_template_path)
        self.shared_urls.append(inp_url)

        build_inp_url = 'file://%s' % (build_inp_path)
        self.shared_urls.append(build_inp_url)

        global_calc_url = 'file://%s' % (global_calc_path)
        self.shared_urls.append(global_calc_url)

        global_calc_url_s = 'file://%s' % (global_calc_path_s)
        self.shared_urls.append(global_calc_url_s)

        ind_calc_url = 'file://%s' % (ind_calc_path)
        self.shared_urls.append(ind_calc_url)

    #---------------------------------------------------------------------------
    #
    def initialize_replicas(self):
        """Initializes replicas and their attributes to default values

           Changed to use geometrical progression for temperature assignment.
        """
        replicas = []
        N = self.replicas
        factor = (self.max_temp/self.min_temp)**(1./(N-1))
        for k in range(N):
            new_temp = self.min_temp * (factor**k)
            r = Replica1d(k, new_temperature=new_temp)
            replicas.append(r)
            
        return replicas

    #---------------------------------------------------------------------------
    #
    def prepare_replica_for_md(self, replica, sd_shared_list):

        basename = self.inp_basename
        template = self.inp_basename[:-5] + ".mdin"
            
        new_input_file = "%s_%d_%d.mdin" % (basename, replica.id, replica.cycle)
        outputname = "%s_%d_%d.mdout" % (basename, replica.id, replica.cycle)
        old_name = "%s_%d_%d" % (basename, replica.id, (replica.cycle-1))
        replica.new_coor = "%s_%d_%d.rst"    % (basename, replica.id, \
                                                replica.cycle)
        replica.new_traj = "%s_%d_%d.mdcrd"  % (basename, replica.id, \
                                                replica.cycle)
        replica.new_info = "%s_%d_%d.mdinfo" % (basename, replica.id, \
                                                replica.cycle)
        replica.old_coor = old_name + ".rst"

        if (replica.cycle == 0):
            first_step = 0
        elif (replica.cycle == 1):
            first_step = int(self.cycle_steps)
        else:
            first_step = (replica.cycle - 1) * int(self.cycle_steps)

        #-----------------------------------------------------------------------

        input_file = "%s_%d_%d.mdin" % (self.inp_basename, \
                                        replica.id, \
                                       (replica.cycle))
        output_file = "%s_%d_%d.mdout" % (self.inp_basename, \
                                          replica.id, \
                                         (replica.cycle))

        new_coor = replica.new_coor
        new_traj = replica.new_traj
        new_info = replica.new_info

        old_coor = replica.old_coor
        old_traj = replica.old_traj

        crds = self.work_dir_local + "/" + \
               self.input_folder + "/" + self.amber_coordinates
        parm = self.work_dir_local + "/" + \
               self.input_folder + "/" + self.amber_parameters

        data = {
            "cycle_steps": str(self.cycle_steps),
            #"new_restraints" : str(self.amber_restraints),
            "new_temperature" : str(replica.new_temperature),
            "amber_input" : str(template),
            "new_input_file" : str(new_input_file),
            "cycle" : str(replica.cycle)
                }
        dump_pre_data = json.dumps(data)
        json_pre_data_bash = dump_pre_data.replace("\\", "")
        json_pre_data_sh = dump_pre_data.replace("\"", "\\\\\"")

        replica.cycle += 1

        #-----------------------------------------------------------------------
        
        stage_out = []
        stage_in = []

        if self.down_mdinfo == True:
            info_local = {
                'source':   new_info,
                'target':   new_info,
                'action':   radical.pilot.TRANSFER
            }
            stage_out.append(info_local)

        if self.down_mdout == True:
            output_local = {
                'source':   output_file,
                'target':   output_file,
                'action':   radical.pilot.TRANSFER
            }
            stage_out.append(output_local)

        info_out = {
            'source': new_info,
            'target': 'staging:///%s' % new_info,
            'action': radical.pilot.COPY
        }
        stage_out.append(info_out)

        coor_out = {
            'source': new_coor,
            'target': 'staging:///%s' % new_coor,
            'action': radical.pilot.COPY
        }
        stage_out.append(coor_out)

        cu = radical.pilot.ComputeUnitDescription()

        if KERNELS[self.resource]["shell"] == "bourne":
            pre_exec_str = "python input_file_builder.py " + "\'" + json_pre_data_sh + "\'"
            cu.executable = '/bin/sh'
        elif KERNELS[self.resource]["shell"] == "bash":
            pre_exec_str = "python input_file_builder.py " + "\'" + json_pre_data_bash + "\'"
            cu.executable = '/bin/bash'          

        if replica.cycle == 1:       
            
            amber_str = self.amber_path
            argument_str = " -O " + " -i " + new_input_file + \
                           " -o " + output_file + \
                           " -p " +  self.amber_parameters + \
                           " -c " + self.amber_coordinates + \
                           " -r " + new_coor + \
                           " -x " + new_traj + \
                           " -inf " + new_info  

            crds_out = {
                'source': self.amber_coordinates,
                'target': 'staging:///%s' % (self.amber_coordinates),
                'action': radical.pilot.COPY
            }
            stage_out.append(crds_out)

            # files needed to be staged in replica dir
            for i in range(4):
                stage_in.append(sd_shared_list[i])
                
            cu.pre_exec = self.pre_exec
            cu.arguments = ['-c', pre_exec_str + "; " + amber_str + argument_str]                                   
            cu.mpi = self.replica_mpi
            cu.cores          = self.replica_cores
            cu.input_staging  = stage_in
            cu.output_staging = stage_out
        else:
            old_coor = "../staging_area/" + self.amber_coordinates

            amber_str = self.amber_path
            argument_str = " -O " + " -i " + new_input_file + \
                           " -o " + output_file + \
                           " -p " +  self.amber_parameters + \
                           " -c " + old_coor + \
                           " -r " + new_coor + \
                           " -x " + new_traj + \
                           " -inf " + new_info

            replica_path = "/replica_%d_%d/" % (replica.id, 0)

            # files needed to be staged in replica dir
            for i in range(4):
                stage_in.append(sd_shared_list[i])

            cu.pre_exec = self.pre_exec
            cu.arguments = ['-c', pre_exec_str + "; " + amber_str + argument_str]   
            cu.mpi = self.replica_mpi
            cu.cores = self.replica_cores
            cu.input_staging = stage_in
            cu.output_staging = stage_out

        return cu

    #---------------------------------------------------------------------------
    #
    def prepare_replica_for_exchange(self, replicas, replica, sd_shared_list):
           
        basename = self.inp_basename
        matrix_col = "matrix_column_%s_%s.dat" % (str(replica.id), str(replica.cycle-1))

        stage_out = []
        col_out = {
            'source': matrix_col,
            'target': 'staging:///%s' % matrix_col,
            'action': radical.pilot.COPY
        }
        stage_out.append(col_out)

        cu = radical.pilot.ComputeUnitDescription()
        cu.executable = "python"
        cu.input_staging  = sd_shared_list[6]
        cu.output_staging = stage_out
        cu.arguments = ["ind_ex_calculator.py", replica.id, (replica.cycle-1), self.replicas, basename, replica.new_temperature]
        cu.cores = 1
        cu.mpi = False
        
        return cu

    #---------------------------------------------------------------------------
    #
    def prepare_global_ex_calc(self, GL, current_cycle, replicas, sd_shared_list):

        stage_out = []
        stage_in = []
        cycle = replicas[0].cycle-1

        outfile = "pairs_for_exchange_{cycle}.dat".format(cycle=cycle)
        stage_out.append(outfile)

        if self.exchange_mpi == True:
            # global_ex_calculator_mpi.py file
            stage_in.append(sd_shared_list[4])

            cu = radical.pilot.ComputeUnitDescription()
            cu.pre_exec = self.pre_exec
            cu.executable = "python"
            cu.input_staging  = stage_in
            cu.arguments = ["global_ex_calculator_mpi.py", str(cycle), str(self.replicas), str(self.inp_basename)]

            # guard for supermic
            if self.replicas > 999:
                self.cores = self.replicas / 2
            elif self.cores < self.replicas:
                if (self.replicas % self.cores) != 0:
                    self.logger.info("Number of replicas must be divisible by the number of Pilot cores!")
                    self.logger.info("pilot cores: {0}; replicas {1}".format(self.cores, self.replicas))
                    sys.exit()
                else:
                    cu.cores = self.cores
            elif self.cores >= self.replicas:
                cu.cores = self.replicas
            else:
                self.logger.info("Number of replicas must be divisible by the number of Pilot cores!")
                self.logger.info("pilot cores: {0}; replicas {1}".format(self.cores, self.replicas))
                sys.exit()
            
            cu.mpi = True         
            cu.output_staging = stage_out
        else:
            # global_ex_calculator.py file
            stage_in.append(sd_shared_list[5])

            cu = radical.pilot.ComputeUnitDescription()
            cu.pre_exec = self.pre_exec
            cu.executable = "python"
            cu.input_staging  = stage_in
            cu.arguments = ["global_ex_calculator.py", str(cycle), str(self.replicas), str(self.inp_basename)]
            cu.cores = 1
            cu.mpi = False         
            cu.output_staging = stage_out
        return cu

    #---------------------------------------------------------------------------
    #
    def perform_swap(self, replica_i, replica_j):
        """Performs an exchange of temperatures

        Arguments:
        replica_i - a replica object
        replica_j - a replica object
        """

        # swap temperatures
        temperature = replica_j.new_temperature
        replica_j.new_temperature = replica_i.new_temperature
        replica_i.new_temperature = temperature
        # record that swap was performed
        replica_i.swap = 1
        replica_j.swap = 1

    #---------------------------------------------------------------------------
    #
    def do_exchange(self, current_cycle, replicas):
        """
        """

        r1 = None
        r2 = None

        cycle = replicas[0].cycle-1

        infile = "pairs_for_exchange_{cycle}.dat".format(cycle=cycle)
        try:
            f = open(infile)
            lines = f.readlines()
            f.close()
            for l in lines:
                pair = l.split()
                r1_id = int(pair[0])
                r2_id = int(pair[1])
                for r in replicas:
                    if r.id == r1_id:
                        r1 = r
                    if r.id == r2_id:
                        r2 = r
                #---------------------------------------------------------------
                # guard
                if r1 == None:
                    rid = random.randint(0,(len(replicas)-1))
                    r1 = replicas[rid]
                if r2 == None:
                    rid = random.randint(0,(len(replicas)-1))
                    r2 = replicas[rid]
                #---------------------------------------------------------------

                # swap parameters
                self.perform_swap(r1, r2)
                r1.swap = 1
                r2.swap = 1
        except:
            raise




"""
.. module:: radical.repex.md_kernles.amber_kernels_salt.kernel_pattern_s_salt
.. moduleauthor::  <haoyuan.chen@rutgers.edu>
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
import time
import math
import json
import random
import shutil
import datetime
from os import path
import radical.pilot
import radical.utils.logger as rul
from kernels.kernels import KERNELS
import amber_kernels_salt.input_file_builder
import amber_kernels_salt.salt_conc_pre_exec
import amber_kernels_salt.salt_conc_post_exec
import amber_kernels_salt.global_ex_calculator

#-------------------------------------------------------------------------------

class KernelPatternSsalt(object):
    
    def __init__(self, inp_file, rconfig,  work_dir_local):
        """Constructor.

        Arguments:
        inp_file - package input file with Pilot and NAMD related parameters as 
        specified by user 
        work_dir_local - directory from which main simulation script was invoked
        """

        self.resource = rconfig['target']['resource']
        self.inp_basename = inp_file['remd.input']['input_file_basename']
        self.inp_folder = inp_file['remd.input']['input_folder']
        self.replicas = int(inp_file['remd.input']['number_of_replicas'])
        self.min_salt = float(inp_file['remd.input']['min_salt'])
        self.max_salt = float(inp_file['remd.input']['max_salt'])
        self.init_temperature = float(inp_file['remd.input']['init_temperature'])
        self.cycle_steps = int(inp_file['remd.input']['steps_per_cycle'])
        self.work_dir_local = work_dir_local
        
        try:
            self.nr_cycles = int(inp_file['remd.input']['number_of_cycles'])
        except:
            self.nr_cycles = None

        if 'exchange_replica_mpi' in inp_file['remd.input']:
            if inp_file['remd.input']['exchange_replica_mpi'] == 'True':
                self.salt_ex_mpi = True
            else:
                self.salt_ex_mpi = False

        if 'replica_mpi' in inp_file['remd.input']:
            mpi = inp_file['remd.input']['replica_mpi']
            if mpi == "True":
                self.replica_mpi = True
            else:
                self.replica_mpi = False
        else:
            self.replica_mpi = False

        if 'download_mdinfo' in inp_file['remd.input']:
            if inp_file['remd.input']['download_mdinfo'] == 'True':
                self.down_mdinfo = True
            else:
                self.down_mdinfo = False
        else:
            self.down_mdinfo = True

        if 'download_mdout' in inp_file['remd.input']:
            if inp_file['remd.input']['download_mdout'] == 'True':
                self.down_mdout = True
            else:
                self.down_mdout = False
        else:
            self.down_mdout = True

        try:
            self.replica_cores = inp_file['remd.input']['replica_cores']
        except:
            self.replica_cores = 1

        self.pre_exec = KERNELS[self.resource]["kernels"]["amber"]["pre_execution"]
        try:
            self.amber_path = inp_file['remd.input']['amber_path']
        except:
            print "Using default Amber path for %s" % rconfig['target']['resource']
            try:
                self.amber_path = KERNELS[self.resource]["kernels"]["amber"]["executable"]
            except:
                print "Amber path for localhost is not defined..."

        self.amber_coordinates = inp_file['remd.input']['amber_coordinates']
        self.amber_parameters = inp_file['remd.input']['amber_parameters']
        self.amber_input = inp_file['remd.input']['amber_input']
        self.input_folder = inp_file['remd.input']['input_folder']

        self.amber_path_mpi = KERNELS[self.resource]["kernels"]["amber"]["executable_mpi"]

        self.shared_urls = []
        self.shared_files = []

        self.name = 'ak-patternB-sc'
        self.ex_name = 'salt-concentration'
        self.logger  = rul.getLogger ('radical.repex', self.name)

    #---------------------------------------------------------------------------
    #
    def prepare_shared_data(self):

        parm_path = self.work_dir_local + "/" + self.input_folder + "/" + \
                    self.amber_parameters
        coor_path = self.work_dir_local + "/" + self.input_folder + "/" + \
                    self.amber_coordinates
                    
        input_template = self.inp_basename[:-5] + ".mdin"
        input_template_path = self.work_dir_local + "/" + \
                              self.input_folder + "/" + \
                              input_template

        build_inp = os.path.dirname(amber_kernels_salt.input_file_builder.__file__)
        build_inp_path = build_inp + "/input_file_builder.py"

        global_calc = os.path.dirname(amber_kernels_salt.global_ex_calculator.__file__)
        global_calc_path = global_calc + "/global_ex_calculator.py"

        salt_pre_exec = os.path.dirname(amber_kernels_salt.salt_conc_pre_exec.__file__)
        salt_pre_exec_path = salt_pre_exec + "/salt_conc_pre_exec.py"

        salt_post_exec = os.path.dirname(amber_kernels_salt.salt_conc_post_exec.__file__)
        salt_post_exec_path = salt_post_exec + "/salt_conc_post_exec.py"
        #-----------------------------------------------------------------------
        self.shared_files.append(self.amber_parameters)
        self.shared_files.append(self.amber_coordinates)
        self.shared_files.append(input_template)
        self.shared_files.append("input_file_builder.py")
        self.shared_files.append("global_ex_calculator.py")
        self.shared_files.append("salt_conc_pre_exec.py")
        self.shared_files.append("salt_conc_post_exec.py")

        parm_url = 'file://%s' % (parm_path)
        self.shared_urls.append(parm_url)  

        coor_url = 'file://%s' % (coor_path)
        self.shared_urls.append(coor_url)   

        inp_url = 'file://%s' % (input_template_path)
        self.shared_urls.append(inp_url)

        build_inp_url = 'file://%s' % (build_inp_path)
        self.shared_urls.append(build_inp_url)

        global_calc_url = 'file://%s' % (global_calc_path)
        self.shared_urls.append(global_calc_url)

        salt_pre_exec_url = 'file://%s' % (salt_pre_exec_path)
        self.shared_urls.append(salt_pre_exec_url)

        salt_post_exec_url = 'file://%s' % (salt_post_exec_path)
        self.shared_urls.append(salt_post_exec_url)

    #---------------------------------------------------------------------------
    #
    def initialize_replicas(self):
        """Initializes replicas and their attributes to default values

           Changed to use geometrical progression for temperature assignment.
        """
        replicas = []
        N = self.replicas
        for k in range(N):
            new_salt = (self.max_salt-self.min_salt)/(N-1)*k + self.min_salt
            r = Replica1d(k, new_salt_concentration=new_salt)
            replicas.append(r)
            
        return replicas

    #---------------------------------------------------------------------------
    # 
    def prepare_replica_for_md(self, replica, sd_shared_list):
        """Prepares all replicas for execution. In this function are created CU 
        descriptions for replicas, are
        specified input/output files to be transferred to/from target system. 
        Note: input files for first and 
        subsequent simulation cycles are different.

        Arguments:
        replicas - list of Replica objects

        Returns:
        compute_replicas - list of radical.pilot.ComputeUnitDescription objects
        """

        basename = self.inp_basename
        template = self.inp_basename[:-5] + ".mdin"
            
        new_input_file = "%s_%d_%d.mdin" % (basename, replica.id, replica.cycle)
        outputname = "%s_%d_%d.mdout" % (basename, replica.id, replica.cycle)
        old_name = "%s_%d_%d" % (basename, replica.id, (replica.cycle-1))
        replica.new_coor = "%s_%d_%d.rst"    % (basename, replica.id, \
                                                replica.cycle)
        replica.new_traj = "%s_%d_%d.mdcrd"  % (basename, replica.id, \
                                                replica.cycle)
        replica.new_info = "%s_%d_%d.mdinfo" % (basename, replica.id, \
                                                replica.cycle)
        replica.old_coor = old_name + ".rst"

        if (replica.cycle == 0):
            first_step = 0
        elif (replica.cycle == 1):
            first_step = int(self.cycle_steps)
        else:
            first_step = (replica.cycle - 1) * int(self.cycle_steps)

        #-----------------------------------------------------------------------

        input_file = "%s_%d_%d.mdin" % (self.inp_basename, \
                                        replica.id, \
                                       (replica.cycle))
        output_file = "%s_%d_%d.mdout" % (self.inp_basename, \
                                          replica.id, \
                                         (replica.cycle))

        new_coor = replica.new_coor
        new_traj = replica.new_traj
        new_info = replica.new_info

        old_coor = replica.old_coor
        old_traj = replica.old_traj

        crds = self.work_dir_local + "/" + \
               self.input_folder + "/" + self.amber_coordinates
        parm = self.work_dir_local + "/" + \
               self.input_folder + "/" + self.amber_parameters

        data = {
            "cycle_steps": str(self.cycle_steps),
            "new_salt_concentration" : str(replica.new_salt_concentration),
            "amber_input" : str(template),
            "new_input_file" : str(new_input_file),
            "cycle" : str(replica.cycle)
                }
        dump_pre_data = json.dumps(data)
        json_pre_data_bash = dump_pre_data.replace("\\", "")
        json_pre_data_sh = dump_pre_data.replace("\"", "\\\\\"")

        replica.cycle += 1

        #-----------------------------------------------------------------------
        
        stage_out = []
        stage_in = []

        if self.down_mdinfo == True:
            info_local = {
                'source':   new_info,
                'target':   new_info,
                'action':   radical.pilot.TRANSFER
            }
            stage_out.append(info_local)

        if self.down_mdout == True:
            output_local = {
                'source':   output_file,
                'target':   output_file,
                'action':   radical.pilot.TRANSFER
            }
            stage_out.append(output_local)


        info_out = {
            'source': new_info,
            'target': 'staging:///%s' % new_info,
            'action': radical.pilot.COPY
        }
        stage_out.append(info_out)

        coor_out = {
            'source': new_coor,
            'target': 'staging:///%s' % new_coor,
            'action': radical.pilot.COPY
        }
        stage_out.append(coor_out)

        cu = radical.pilot.ComputeUnitDescription()

        if KERNELS[self.resource]["shell"] == "bourne":
            pre_exec_str = "python input_file_builder.py " + "\'" + json_pre_data_sh + "\'"
            cu.executable = '/bin/sh'
        elif KERNELS[self.resource]["shell"] == "bash":
            pre_exec_str = "python input_file_builder.py " + "\'" + json_pre_data_bash + "\'"
            cu.executable = '/bin/bash'          

        print self.replica_mpi

        if replica.cycle == 1:       
            
            amber_str = self.amber_path
            argument_str = " -O " + " -i " + new_input_file + \
                           " -o " + output_file + \
                           " -p " +  self.amber_parameters + \
                           " -c " + self.amber_coordinates + \
                           " -r " + new_coor + \
                           " -x " + new_traj + \
                           " -inf " + new_info  

            crds_out = {
                'source': self.amber_coordinates,
                'target': 'staging:///%s' % (self.amber_coordinates),
                'action': radical.pilot.COPY
            }
            stage_out.append(crds_out)

            # files needed to be staged in replica dir
            for i in range(4):
                stage_in.append(sd_shared_list[i])
                
            cu.pre_exec = self.pre_exec
            cu.arguments = ['-c', pre_exec_str + "; " + amber_str + argument_str]                                   
            cu.mpi = self.replica_mpi
            cu.cores          = self.replica_cores
            cu.input_staging  = stage_in
            cu.output_staging = stage_out
        else:
            old_coor = "../staging_area/" + self.amber_coordinates

            amber_str = self.amber_path
            argument_str = " -O " + " -i " + new_input_file + \
                           " -o " + output_file + \
                           " -p " +  self.amber_parameters + \
                           " -c " + old_coor + \
                           " -r " + new_coor + \
                           " -x " + new_traj + \
                           " -inf " + new_info

            replica_path = "/replica_%d_%d/" % (replica.id, 0)

            # files needed to be staged in replica dir
            for i in range(4):
                stage_in.append(sd_shared_list[i])

            cu.pre_exec = self.pre_exec
            cu.arguments = ['-c', pre_exec_str + "; " + amber_str + argument_str]   
            cu.mpi = self.replica_mpi
            cu.cores = self.replica_cores
            cu.input_staging = stage_in
            cu.output_staging = stage_out

        return cu

    #---------------------------------------------------------------------------
    # 
    def prepare_replica_for_exchange(self, replicas, replica, sd_shared_list):
        """Creates a list of ComputeUnitDescription objects for exchange step 
        on resource.
        Number of matrix_calculator_s2.py instances invoked on resource is equal 
        to the number 
        of replicas. 

        Arguments:
        replicas - list of Replica objects

        Returns:
        exchange_replicas - list of radical.pilot.ComputeUnitDescription objects
        """

        # name of the file which contains swap matrix column data for each replica
        basename = self.inp_basename
        matrix_col = "matrix_column_%s_%s.dat" % (str(replica.id), str(replica.cycle-1))

        all_salts = {}
        for r in replicas:
            all_salts[str(r.id)] = str(r.new_salt_concentration)

        cu = radical.pilot.ComputeUnitDescription()
        
        data = {
            "replica_id": str(replica.id),
            "replica_cycle" : str(replica.cycle-1),
            "replicas" : str(self.replicas),
            "base_name" : str(basename),
            "init_temp" : str(self.init_temperature),
            "init_salt" : str(replica.new_salt_concentration),
            "amber_path" : str(self.amber_path),
            "amber_input" : str(self.amber_input),
            "all_salts" : all_salts, 
            "amber_parameters": str(self.amber_parameters), 
            "r_old_path": str(replica.old_path),
        }

        dump_data = json.dumps(data)
        json_data = dump_data.replace("\\", "")

        salt_pre_exec = ["python salt_conc_pre_exec.py " + "\'" + json_data + "\'"]
        cu.pre_exec = self.pre_exec + salt_pre_exec
        cu.executable = self.amber_path_mpi
        salt_post_exec = ["python salt_conc_post_exec.py " + "\'" + json_data + "\'"]
        cu.post_exec = salt_post_exec

        rid = replica.id
        in_list = []
        in_list.append(sd_shared_list[0])
        in_list.append(sd_shared_list[2])
        in_list.append(sd_shared_list[5])
        in_list.append(sd_shared_list[6])
  
        cu.input_staging = in_list

        out_list = []
        matrix_col_out = {
            'source': matrix_col,
            'target': 'staging:///%s' % (matrix_col),
            'action': radical.pilot.COPY
        }
        out_list.append(matrix_col_out)

        cu.arguments = ['-ng', str(self.replicas), '-groupfile', 'groupfile']
        cu.cores = self.replicas
        cu.mpi = self.salt_ex_mpi
        cu.output_staging = out_list   

        return cu

    #---------------------------------------------------------------------------
    #
    def prepare_global_ex_calc(self, GL, current_cycle, replicas, sd_shared_list):

        stage_out = []
        stage_in = []

        if GL == 1:
            cycle = replicas[0].cycle-1
        else:
            cycle = replicas[0].cycle
        
        # global_ex_calculator.py file
        stage_in.append(sd_shared_list[4])

        outfile = "pairs_for_exchange_{cycle}.dat".format(cycle=cycle)
        stage_out.append(outfile)

        cu = radical.pilot.ComputeUnitDescription()
        cu.pre_exec = self.pre_exec
        cu.executable = "python"
        cu.input_staging  = stage_in
        cu.arguments = ["global_ex_calculator.py", str(self.replicas), str(cycle)]
        cu.cores = 1
        cu.mpi = False            
        cu.output_staging = stage_out

        return cu

    #---------------------------------------------------------------------------
    #
    def perform_swap(self, replica_i, replica_j):
        """Performs an exchange of salt concentrations

        Arguments:
        replica_i - a replica object
        replica_j - a replica object
        """

        # swap temperatures
        temp = replica_j.new_salt_concentration
        replica_j.new_salt_concentration = replica_i.new_salt_concentration
        replica_i.new_salt_concentration = temp
        # record that swap was performed
        replica_i.swap = 1
        replica_j.swap = 1

    #---------------------------------------------------------------------------
    #
    def do_exchange(self, current_cycle, replicas):
        """
        """

        r1 = None
        r2 = None

        cycle = replicas[0].cycle-1

        infile = "pairs_for_exchange_{cycle}.dat".format(cycle=cycle)
        try:
            f = open(infile)
            lines = f.readlines()
            f.close()
            for l in lines:
                pair = l.split()
                r1_id = int(pair[0])
                r2_id = int(pair[1])
                for r in replicas:
                    if r.id == r1_id:
                        r1 = r
                    if r.id == r2_id:
                        r2 = r
                #---------------------------------------------------------------
                # guard
                if r1 == None:
                    rid = random.randint(0,(len(replicas)-1))
                    r1 = replicas[rid]
                if r2 == None:
                    rid = random.randint(0,(len(replicas)-1))
                    r2 = replicas[rid]
                #---------------------------------------------------------------

                # swap parameters
                self.perform_swap(r1, r2)
                r1.swap = 1
                r2.swap = 1
        except:
            raise




""" Application kernel configuration file.
"""

KERNELS = {

    "xsede.stampede":
    {
        "params":
        {
            "cores": 16,
        },
        "kernels":
        {
            "amber": {
                "environment" : {},
                "pre_execution" : ["module restore", "module load amber", "module load python"],
                "executable" : "/opt/apps/intel13/mvapich2_1_9/amber/12.0/bin/sander",
                "executable_mpi" : "/opt/apps/intel13/mvapich2_1_9/amber/12.0/bin/sander.MPI"

            },
            "namd": {
                "environment" : {},
                "pre_execution" : ["module load TACC", "module load namd/2.9"],
                "executable" : "/opt/apps/intel13/mvapich2_1_9/namd/2.9/bin/namd2"
            }
        },
        "shell": "bash"
    },"epsrc.archer":
    {
        "params":
        {
            "cores": 24,
        },
        "kernels":
        {
            "amber": {
                "environment" : {},
                "pre_execution" : ["python-compute/2.7.6", "module load amber", "module use --append /work/e290/e290/marksant/privatemodules", "module load openmpi/HEAD", "module switch PrgEnv-cray PrgEnv-gnu"],
                "executable" : "/work/y07/y07/amber/12/bin/sander",
                "executable_mpi" : "/work/y07/y07/amber/12/bin/sander.MPI"
            }
        },
        "shell": "bash"
    },"xsede.gordon":
    {
        "params":
        {
            "cores": 16,
        },
        "kernels":
        {
            "namd": {
                "environment" : {},
                "pre_execution" : ["module load namd/2.9"],
                "executable" : "/opt/namd/bin/namd2"
            },
            "amber": {
                "environment" : {},
                "pre_execution" : ["module load python", "module load amber"],
                "executable" : "/opt/amber/bin/sander",
                "executable_mpi" : "/opt/amber/bin/sander.MPI"
            }
        },
        "shell": "bash"
    },"local.localhost":
    {
        "params":
        {
            "cores": 1,
        },
        "kernels":
        {
            "amber": {
                "pre_execution" : [],
                "executable" : "/home/antons/amber/amber14/bin/sander",
                "executable_mpi" : "/home/antons/amber/amber14/bin/sander"
            }
        },
        "shell": "bash"
    },"xsede.comet":
    {
        "params":
        {
            "cores": 24,
        },
        "kernels":
        {
            "amber": {
                "environment" : {},
                "pre_execution" : ["module load python", "module load mpi4py/1.3.1", "module load amber"],
                "executable" : "/opt/amber/bin/sander",
                "executable_mpi" : "/opt/amber/bin/sander.MPI"
            }
        },
        "shell": "bourne"
    },"ncsa.bw":
    {
        "params":
        {
            "cores": 32,
        },
        "kernels":
        {
            "amber": {
                "environment" : {},
                "pre_execution" : [],
                "executable" : "/u/sciteam/treikali/amber14/bin/sander",
                "executable_mpi" : "/u/sciteam/treikali/amber14/bin/sander.MPI"
            }
        }
    },"xsede.supermic":
    {
        "params":
        {
            "cores": 20,
        },
        "kernels":
        {
            "amber": {
                "environment" : {},
                "pre_execution" : ["module unload python/2.7.7-anaconda", "module load python/2.7.7/GCC-4.9.0", "module load amber/14/INTEL-140-MVAPICH2-2.0"],
                "executable" : "/home/antontre/amber14/bin/sander",
                "executable_mpi" : "/usr/local/packages/amber/14/INTEL-140-MVAPICH2-2.0/bin/sander.MPI"
            }
        },
        "shell": "bash"
    }

}



"""
.. module:: radical.repex.namd_kernels.namd_kernel
.. moduleauthor::  <antons.treikalis@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__ = "MIT"

import os
import sys
from os import path
import radical.pilot
from kernels.kernels import KERNELS
from md_kernels.md_kernel_tex import *

#-----------------------------------------------------------------------------------------------------------------------------------

class NamdKernelTex(MdKernelTex):
    """
    """
    def __init__(self, inp_file,  work_dir_local):
        """Constructor.

        Arguments:
        inp_file - package input file with Pilot and NAMD related parameters as specified by user 
        work_dir_local - directory from which main simulation script was invoked
        """

        MdKernelTex.__init__(self, inp_file, work_dir_local)

        try:
            self.namd_path = inp_file['input.MD']['namd_path']
        except:
            print "Using default NAMD path for %s" % inp_file['input.PILOT']['resource']
            try:
                self.namd_path = KERNELS[self.resource]["kernels"]["namd"]["executable"]
            except:
                print "NAMD path for localhost is not defined..."
        
        self.namd_structure = inp_file['input.MD']['namd_structure']
        self.namd_coordinates = inp_file['input.MD']['namd_coordinates']
        self.namd_parameters = inp_file['input.MD']['namd_parameters']
        
#----------------------------------------------------------------------------------------------------------------------------------

    def build_input_file_local(self, replica):
        """Generates input file for individual replica, based on template input file. Tokens @xxx@ are
        substituted with corresponding parameters. 
        In this function replica.cycle is incremented by one

        old_output_root @oldname@ determines which .coor .vel and .xsc files are used for next cycle

        Arguments:
        replica - a single Replica object
        """

        basename = self.inp_basename[:-5]
        template = self.inp_basename
            
        new_input_file = "%s_%d_%d.namd" % (basename, replica.id, replica.cycle)
        outputname = "%s_%d_%d" % (basename, replica.id, replica.cycle)
        old_name = "%s_%d_%d" % (basename, replica.id, (replica.cycle-1))
        replica.new_coor = outputname + ".coor"
        replica.new_vel = outputname + ".vel"
        replica.new_history = outputname + ".history"
        replica.new_ext_system = outputname + ".xsc" 
        historyname = replica.new_history

        replica.old_coor = old_name + ".coor"
        replica.old_vel = old_name + ".vel"
        replica.old_ext_system = old_name + ".xsc" 

        if (replica.cycle == 0):
            first_step = 0
        elif (replica.cycle == 1):
            first_step = int(self.cycle_steps)
        else:
            first_step = (replica.cycle - 1) * int(self.cycle_steps)

        old_name = "%s_%d_%d" % (basename, replica.id, (replica.cycle-1)) 
        structure = self.namd_structure
        coordinates = self.namd_coordinates
        parameters = self.namd_parameters

        # substituting tokens in main replica input file 
        try:
            r_file = open( (os.path.join((self.work_dir_local + "/namd_inp/"), template)), "r")
        except IOError:
            print 'Warning: unable to access template file %s' % template

        tbuffer = r_file.read()
        r_file.close()

        tbuffer = tbuffer.replace("@swap@",str(replica.swap))
        tbuffer = tbuffer.replace("@ot@",str(replica.old_temperature))
        tbuffer = tbuffer.replace("@nt@",str(replica.new_temperature))
        tbuffer = tbuffer.replace("@steps@",str(self.cycle_steps))
        tbuffer = tbuffer.replace("@rid@",str(replica.id))
        tbuffer = tbuffer.replace("@somename@",str(outputname))
        tbuffer = tbuffer.replace("@oldname@",str(old_name))
        tbuffer = tbuffer.replace("@cycle@",str(replica.cycle))
        tbuffer = tbuffer.replace("@firststep@",str(first_step))
        tbuffer = tbuffer.replace("@history@",str(historyname))
        tbuffer = tbuffer.replace("@structure@", str(structure))
        tbuffer = tbuffer.replace("@coordinates@", str(coordinates))
        tbuffer = tbuffer.replace("@parameters@", str(parameters))
        
        replica.cycle += 1
        # write out
        try:
            w_file = open( new_input_file, "w")
            w_file.write(tbuffer)
            w_file.close()
        except IOError:
            print 'Warning: unable to access file %s' % new_input_file

#-----------------------------------------------------------------------------------------------------------------------------------

    def prepare_replicas_local(self, replicas):
        """Creates a list of ComputeUnitDescription objects for MD simulation step. Here are
        specified input/output files to be transferred to/from target resource. Note: input 
        files for first and subsequent simulaition cycles are different.  

        Arguments:
        replicas - list of Replica objects

        Returns:
        compute_replicas - list of radical.pilot.ComputeUnitDescription objects
        """
        compute_replicas = []
        for r in range(len(replicas)):
            self.build_input_file_local(replicas[r])
            input_file = "%s_%d_%d.namd" % (self.inp_basename[:-5], replicas[r].id, (replicas[r].cycle-1))

            new_coor = replicas[r].new_coor
            new_vel = replicas[r].new_vel
            new_history = replicas[r].new_history
            new_ext_system = replicas[r].new_ext_system

            old_coor = replicas[r].old_coor
            old_vel = replicas[r].old_vel
            old_ext_system = replicas[r].old_ext_system 

            # only for first cycle we transfer structure, coordinates and parameters files
            if replicas[r].cycle == 1:
                cu = radical.pilot.ComputeUnitDescription()
                cu.pre_exec    = KERNELS[self.resource]["kernels"]["namd"]["pre_execution"]
                cu.executable = self.namd_path
                cu.arguments = [input_file]
                cu.cores = replicas[r].cores
                structure = self.work_dir_local + "/" + self.inp_folder + "/" + self.namd_structure
                coords = self.work_dir_local + "/" + self.inp_folder + "/" + self.namd_coordinates
                params = self.work_dir_local + "/" + self.inp_folder + "/" + self.namd_parameters
                cu.input_staging = [str(input_file), str(structure), str(coords), str(params)]
                cu.output_staging = [str(new_coor), str(new_vel), str(new_history), str(new_ext_system) ]
                compute_replicas.append(cu)
            else:
                cu = radical.pilot.ComputeUnitDescription()
                cu.pre_exec    = KERNELS[self.resource]["kernels"]["namd"]["pre_execution"]
                cu.executable = self.namd_path
                cu.arguments = [input_file]
                cu.cores = 1
                structure = self.inp_folder + "/" + self.namd_structure
                coords = self.inp_folder + "/" + self.namd_coordinates
                params = self.inp_folder + "/" + self.namd_parameters
                cu.input_staging = [str(input_file), str(structure), str(coords), str(params), str(old_coor), str(old_vel), str(old_ext_system)]
                cu.output_staging = [str(new_coor), str(new_vel), str(new_history), str(new_ext_system) ]
                compute_replicas.append(cu)

        return compute_replicas

#----------------------------------------------------------------------------------------------------------------------------------

    def get_historical_data(self, replica, cycle):
        """Retrieves temperature and potential energy from simulaion output file <file_name>.history
        """
        if not os.path.exists(replica.new_history):
            print "history file not found: "
            print replica.new_history
        else:
            f = open(replica.new_history)
            lines = f.readlines()
            f.close()
            data = lines[0].split()
         
        return float(data[0]), float(data[1])